#!/usr/bin/env python

# ----------------------------------------------------------------------
# Numenta Platform for Intelligent Computing (NuPIC)
# Copyright (C) 2013, Numenta, Inc.  Unless you have an agreement
# with Numenta, Inc., for a separate license for this software code, the
# following terms and conditions apply:
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero Public License for more details.
#
# You should have received a copy of the GNU Affero Public License
# along with this program.  If not, see http://www.gnu.org/licenses.
#
# http://numenta.org/licenses/
# ----------------------------------------------------------------------

"""
## @file
SVMClassifierNode implements Support Vector Machines (SVMs)
"""

import math
import random
import logging

import numpy
from nupic.pynodes.PyNode import (PyNode,
                                  NodeSpec,
                                  NodeSpecItem,
                                  RealTypeName,
                                  RealNumpyDType)
from nupic.pynodes import PyNode as PyNodeModule
from nupic.algorithms import svm_01, svm_dense
from nupic.analysis.memoryAwareness import MemoryAwareness


_kKernelTypes = ["rbf", "linear"]



class SVMClassifierNode(PyNode, MemoryAwareness):
  """
  SVMClassifierNode implements Support Vector Machines (SVMs), which can be used to
  perform supervised learning by mapping a set of top-level groups beliefs onto
  a set of category labels. The node is a wrapper around a modified version
  of the libsvm library.
  """

  def __init__(self,
      categoriesOut=2,
      # SVM parameter ranges
      minC=0.0,
      maxC=0.0,
      minGamma=0.0,
      maxGamma=0.0,
      kernelType='rbf',
      # Latin Hypercube sampling (LHS)
      numSamplesPerRecursion=1,
      numRecursions=0,
      contractionFactor=0.3,
      numCrossValidations=5,
      # Implementation
      convEpsilon=0.01,
      useSparseSvm=False,
      inputThresh=0.500,
      useProbabilisticSvm=True,
      doSphering=False,
      deterministic=False,
      nta_cpp_svm_seed=-1,
      # PCA
      numSVDSamples=None,
      numSVDDims=None,
      fractionOfMax=None,
      useAuxiliary=False,
      justUseAuxiliary=False,
      discardProblem=True,  # Set to False to keep data for PCA visualizer
      # KNN-type stuff
      keepSamples=False,  # Keep the original inputs around
      calculateDistances=False,  # Calculate and store distances to orig. inputs
      nta_monitorMemory=False,
    ):
    """
    @param categoriesOut -- The maximum number of distinct category
          labels that can be learned.
    """
    MemoryAwareness.__init__(self)
    self.clear()
    self._firstComputeCall = True
    self._inputVector = None
    self._scanInfo = None
    self._scanResults = None

    # SVM parameters
    self.C = minC
    self.minC = minC
    self.maxC = maxC
    self.gamma = minGamma
    self.minGamma = minGamma
    self.maxGamma = maxGamma
    self._kernelType = kernelType

    # Latin Hypercube Sampling
    self.numSamplesPerRecursion = numSamplesPerRecursion
    self.numRecursions = numRecursions
    self.contractionFactor = contractionFactor
    self.numCrossValidations = numCrossValidations

    # Bindings parameters
    self.convEpsilon = convEpsilon
    self.useSparseSvm = useSparseSvm
    self.inputThresh = inputThresh
    self.useProbabilisticSvm = useProbabilisticSvm
    self.doSphering = doSphering
    self.deterministic = deterministic
    self.cpp_svm_seed = nta_cpp_svm_seed

    # PCA
    self._numSVDSamples = numSVDSamples
    self._numSVDDims = numSVDDims
    self._fractionOfMax = fractionOfMax
    self._useAuxiliary = useAuxiliary
    self._justUseAuxiliary = justUseAuxiliary
    self._auxInputLen = None
    if numSVDDims=='adaptive':
      self._adaptiveSVDDims = True
    else:
      self._adaptiveSVDDims = False
    self.discardProblem = discardProblem

    # KNN-type stuff
    self.keepSamples = keepSamples
    self.calculateDistances = calculateDistances

    # Memory monitoring
    self._enableMonitoring(nta_monitorMemory)
    self._initRandom()
    self.clear()

  def clear(self):
    """Clear all persistent internal state."""

    self._learningMode = True
    self._inferenceMode = False
    self._autoTuningData = False
    self._inputWidth = None
    self._catIdMap = None
    self._svm = None
    self._svmParams = None
    self._samples = None
    self._labels  = None
    self._partitionIds = None
    self._upcomingPartitionIds = None
    # Support for using non-training samples for
    # (C, Gamma) optimization
    self._autoTuneSamples      = None
    self._autoTuneLabels       = None
    self._autoTunePartitionIds = None

    # We operate in two modes:
    #   'classification' - the SVM is used for top-level classification
    #   'feedback' - the SVM is a "temporary" classified whose real purpose
    #                is simply to identify the most useful features
    #                so that this information can be fed back to the
    #                feature selection stage for pruning purposes.
    # The only difference is the choice of kernel; for 'classification'
    # mode, then kernel is selected based on the value of the parameter
    # 'kernelType'.  But in 'feedback' mode, the kernel is always
    # of type 'linear' regardless of the value of 'kernelType'.
    self._mode = 'classification'

    # Sphering normalization
    self._normOffset = None
    self._normScale  = None

    # PCA
    self._numPatterns = 0
    self._s = None
    self._vt = None
    self._mean = None

    # KNN-type stuff
    self.distances = None
    self._distanceCount = 0


  def _initRandom(self):
    """
    Create and seed random number generator.
    """
    # Create PRNG
    self._rng = random.Random()
    # Seed
    if self.deterministic:
      self._rng.seed(42)
    else:
      self._rng.seed()

  def _setInferenceMode(self, parameterValue):
    # If inference mode is being turned on, build the SVM model and turn off learning mode
    # This node only supports a one-time switch from learning mode to inference mode
    value = bool(int(parameterValue))
    if self._learningMode and value:
      self._finishLearning()
    assert self._inferenceMode == value

  inferenceMode = property(
      fget=lambda self: self._inferenceMode,
      fset=_setInferenceMode,
      doc="""Boolean indicating whether or not a node
          is in inference mode"""
    )

  def _setLearningMode(self, parameterValue):
    # If learning mode is being turned off, build the SVM model and turn on inference mode
    # This node only supports a one-time switch from learning mode to inference mode
    value = bool(int(parameterValue))
    if self._learningMode and not value:
      self._finishLearning()
    assert self._learningMode == value

  learningMode = property(
      fget=lambda self: self._learningMode,
      fset=_setLearningMode,
      doc="""Boolean indicating whether or not a node
          is in learning mode"""
    )

  def _setMode(self, parameterValue):
    # We can only change this before we have received our first compute()
    # call.  Otherwise it is a runtime error.
    if self._svm is not None:
      raise RuntimeError, "SVMClassifierNode 'mode' parameter cannot be changed" \
                          "after first compute() call"
    self._mode = parameterValue

  mode = property(
      fget=lambda self: self._mode,
      fset=_setMode,
      doc="""We operate in two modes: 'classification' - the SVM is
          used for top-level classification; 'feedback' - the SVM is a "temporary"
          classified whose real purpose is simply to identify the most useful
          features so that this information can be fed back to the feature
          selection stage for pruning purposes.  The only difference is the
          choice of kernel; for 'classification' mode, then kernel is selected
          based on the value of the parameter 'kernelType'.  But in 'feedback'
          mode, the kernel is always of type 'linear' regardless of the value
          of 'kernelType'."""
    )

  def _setKernelType(self, parameterValue):
    # Convert in case they passed an integer, since technically we are an
    # enum...
    try:    parameterValue = _kKernelTypes[parameterValue]
    except: pass
    self._kernelType = parameterValue

  kernelType = property(
      fget=lambda self: self._kernelType,
      fset=_setKernelType,
      # Doc provided with constraints in NodeSpec.
    )

  def _set_nta_cpp_svm_seed(self, parameterValue):
    self.cpp_svm_seed = parameterValue
  nta_cpp_svm_seed = property(fget=lambda self: self.cpp_svm_seed,
                              fset=_set_nta_cpp_svm_seed)


  def compute(self, nodeInfo, inputs, outputs):
    """
    Process one input sample.
    This method is called by the runtime engine.
    """

    # If the first time being called, then print potential warning messsages
    if self._firstComputeCall:
      self._firstComputeCall = False
      if self._useAuxiliary:
#        print "\n  Auxiliary input stream from Image Sensor enabled."
        if self._justUseAuxiliary == True:
          print "  Warning: You have chosen to ignore the image data and instead just use the auxiliary data stream."

    if self._scanInfo is not None:
      # The input is much larger than what we've been trained on
      # Sweep across it, assembling the outputs in a data structure
      self._scan(inputs)
      return

    # Assemble inputs
    childInputs = [x.wvector(0) for x in inputs['bottomUpIn']]
    inputVector = numpy.concatenate([x.array() for x in childInputs])

#    # Look for auxiliary input
#    if self._useAuxiliary==True:
#      auxVector = inputs['auxDataIn'][0].wvector(0).array()
#      if auxVector.dtype != numpy.float32:
#        raise RuntimeError, "SVMClassifierNode expects numpy.float32 for the auxiliary data vector"
#      if self._justUseAuxiliary == True:
#        inputVector = inputs['auxDataIn'][0].wvector(0).array()
#      else:
#        inputVector = numpy.concatenate([inputVector, inputs['auxDataIn'][0].wvector(0).array()])

    # Look for auxiliary input
    if self._useAuxiliary:
      auxVector = inputs['auxDataIn'][0].wvector(0).array()
      if auxVector.dtype != numpy.float32:
        raise RuntimeError, "SVMClassifierNode expects numpy.float32 for the auxiliary data vector"
      if self._justUseAuxiliary == True:
        inputVector = auxVector
      else:
        inputVector = numpy.concatenate([inputVector, auxVector])
    else:
      auxVector = numpy.array([])

    if self._auxInputLen == None:
      self._auxInputLen = len(auxVector)

    # Initialize data structures the first time
    if self._inputWidth is None:
      self._initDataStructures(len(inputVector))

    # Learn
    if self._learningMode:
      # Extract category label
      catInput = int(inputs["categoryIn"][0].wvector(0)[0])

      # Project the vector onto the PCA basis if present
      if self._vt is not None:
        inputVector = numpy.dot(self._vt, inputVector - self._mean)
        if self._useAuxiliary and not self._justUseAuxiliary:
          inputVector = numpy.concatenate([inputVector, auxVector])

      if self._upcomingPartitionIds:
        partitionId = self._upcomingPartitionIds.pop(0)
      else:
        partitionId = None

      # Perform learning
      self._learn(inputVector, catInput, partitionId)

      # Perform PCA if it is time to do so
      if self._vt is None and self._numSVDDims is not None \
          and self._numSVDSamples is not None \
          and self._samples.shape[0] == self._numSVDSamples:
        self.computeSVD()

    # Inference
    elif self._inferenceMode:
      # Project the vector onto the PCA basis if present
      if self._vt is not None:
        if self._useAuxiliary and not self._justUseAuxiliary:
          auxVector = inputVector[len(inputVector)-self._auxInputLen:]
          inputVector = inputVector[:len(inputVector)-self._auxInputLen]
          inputVector = numpy.dot(self._vt, inputVector - self._mean)
          inputVector = numpy.concatenate([inputVector, auxVector])
        else:
          inputVector = numpy.dot(self._vt, inputVector - self._mean)

      # Save the input vector, in case the distances are requested later
      self._inputVector = inputVector

      allOutputs = outputs['categoriesOut'].wvector()
      inferenceResult = self._infer(inputVector)
      allOutputs.fill(0)

      # Convert from internal indices back to ImageSensor's IDs
      # This is necessary to calculate accuracy by comparing our outputs to
      # the true category from the sensor
      remap = [self.catIndexToId(i) for i in xrange(len(self._catIdMap))]
      out = numpy.zeros(max(remap)+1)
      out[remap] = inferenceResult[0:len(self._catIdMap)]
      nout = min(len(allOutputs), len(out))
      allOutputs[0:nout] = out[0:nout]


  def _scan(self, inputs):
    """
    Run scanning inference and store the results.

    The input is from many nodes, but we were trained with just a single child.
    Perform inference on each node separately and store in a list.
    """

    childInputs = [x.wvector(0).array() for x in inputs['bottomUpIn']]
    self._scanResults = []

    for inputVector in childInputs:

      # Project the vector onto the PCA basis if present
      if self._vt is not None:
        inputVector = numpy.dot(self._vt, inputVector - self._mean)

      # Run inference
      inferenceResult = self._infer(inputVector)

      # Convert from internal indices back to ImageSensor's IDs
      # This is necessary to calculate accuracy by comparing our outputs to
      # the true category from the sensor
      remap = [self.catIndexToId(i) for i in xrange(len(self._catIdMap))]
      out = numpy.zeros(max(remap)+1)
      out[remap] = inferenceResult[0:len(self._catIdMap)]

      # Store
      self._scanResults.append(tuple(out))


  # @todo -- Modernize nodeSpec
  def getNodeSpec(self):
    """
    Return the NodeSpec for this PyNode.
    """

    parent = PyNode.getNodeSpec(self)
    out = NodeSpec(
        description=SVMClassifierNode.__doc__,
        inputs=[
          NodeSpecItem(name="categoryIn", type=RealTypeName,
                       description="""Category of the input sample"""),
          NodeSpecItem(name="bottomUpIn", type=RealTypeName,
                       description="""Belief values over children's groups"""),
           NodeSpecItem(name="auxDataIn", type=RealTypeName,
                        description="""Auxiliary data from the sensor""")
          ],
        outputs=[
          NodeSpecItem(name="categoriesOut", type=RealTypeName,
                      description="A vector representing, for each category" \
                                   " index, the likelihood that the input to" \
                                   " the node belongs to that category.")
          ],
        parameters=[
          NodeSpecItem(name="learningMode", type="bool", constraints="bool", access="gs", value = True,
                       description="Boolean indicating whether or not a node " \
                                   "is in learning mode"),
          NodeSpecItem(name="inferenceMode", type="bool", constraints="bool", access="gs", value = False,
                       description="Boolean indicating whether or not a node " \
                                   "is in inference mode"),
          NodeSpecItem(name="activeOutputCount", type="uint", access="g",
                       description="The number of active elements in the "    \
                                   "'categoriesOut' output."),
          NodeSpecItem(name="categoryCount", type="uint", access="g",
                       description="An integer indicating the number of "    \
                                   "categories that have been learned"),
          NodeSpecItem(name="C", type="float", access="g", value = 0.0,
                       description="""The current value of C, an SVM parameter that
                          influences the error rate. If numRecursions==0, this value is
                          used to build the SVM; otherwise, minC and maxC are used,
                          and this parameter is changed to the best value afterwards."""),
          NodeSpecItem(name="minC", type="float", access="cg", value = 0.0,
                       description="""The minimum value of C, an SVM parameter that
                          influences the error rate. SVMClassifierNode will
                          perform an optimization process in order to find the
                          value of C that minimizes error
                          rate on the set of training samples."""),
          NodeSpecItem(name="maxC", type="float", access="cg", value = 0.0,
                       description="""The maximum value of C, an SVM parameter that
                          influences the error rate."""),
          NodeSpecItem(name="gamma", type="float", access="g", value = 0.0,
                       description="""The current value of gamma, an SVM parameter that
                          influences the error rate. If numRecursions==0, this value is
                          used to build the SVM; otherwise, minGamma and maxGamma are used,
                          and this parameter is changed to the best value afterwards."""),
          NodeSpecItem(name="minGamma", type="float", access="cg", value = 0.0,
                       description="""The minimum value of Gamma, an SVM parameter that
                          influences the error rate.  SVMClassifierNode will perform an
                          optimization process in order to find the value of Gamma that
                          minimizes error rate on the set of training samples."""),
          NodeSpecItem(name="maxGamma", type="float", access="cg", value = 0.0,
                       description="""The maximum value of Gamma, an SVM parameter that
                          influences the error rate."""),
          NodeSpecItem(name="kernelType", type="string", access="cgs", value= "rbf",
                       description="""Specifies the type of kernel to use.  Valid choices
                          are: 'rbf' (for radial basis function kernel), or 'linear' (for
                          linear kernel.)  Default is 'rbf'.""",
                          constraints="enum: %s,%s" % (
                                        ",".join(_kKernelTypes),
                                        ",".join(map(str, xrange(len(_kKernelTypes))))
                                      )),
          NodeSpecItem(name="numRecursions", type="uint", access="cgs", value = 0,
                       description="""The number of rounds of recursive Latin Hypercube Sampling
                          to perform.  If set to 0 (the default), then no parameter search is
                          performed, and instead the values 'minC' and 'minGamma' are used."""),
          NodeSpecItem(name="numSamplesPerRecursion", type="uint", access="gs", value = 1,
                       description="""The number of samples we will test in each round of
                          recursive Latin Hypercube Sampling."""),
          NodeSpecItem(name="contractionFactor", type="float", access="cgs", value = 0.3,
                       description="""The fraction of the C and Gamma sampling space that we will
                          sample with each successive round of Latin Hypercube sampling."""),
          NodeSpecItem(name="numCrossValidations", type="uint", access="cg", value = 5,
                       description="""The number of cross validation steps
                          steps that are used to estimate the error rate for a given
                          set of C and Gamma values.  Increasing this number will yield
                          better error rate estimates, and hence more optimal values
                          for C and Gamma, but will take longer to complete the
                          learning process."""),
          NodeSpecItem(name="autoTuningData", type="bool", constraints="bool", access="gs",
                       description="""Data presented to the classifier while autoTuningData is set to True
                          will be used as a test set for the auto-tuning phase and is NOT
                          learned by the node. The auto-tuning phase is performed at the
                          end of learning (when inferenceMode is first set to True)."""),
          NodeSpecItem(name="convEpsilon", type="float", access="cg", value = 0.01,
                       description="""A parameter that controls the convergence
                          criterion used to halt the search for optimal SVM hyperplanes."""),
          NodeSpecItem(name="useSparseSvm", type="bool", constraints="bool", access="gc", value = False,
                       description="""A boolean that controls whether input vectors
                          should be binarized during learning and inference."""),
          NodeSpecItem(name="inputThresh", type="float", access="cg", value = 0.5,
                       description="""A floating point value that establishes the threshold
                          used for binarizing input vectors during learning and inference.
                          If 'useSparseSvm' is False, then this parameter has no effect."""),
          NodeSpecItem(name="useProbabilisticSvm", type="bool", constraints="bool", access="cg", value = True,
                       description="""A boolean that controls whether or not to
                          build an underlying SVM model that is capable of estimating
                          probabilistic beliefs during inference, as opposed to simply
                          determining a single winner-take-all category."""),
          NodeSpecItem(name="doSphering", type="bool", constraints="bool", access="cg", value = False,
                       description="""A boolean indicating whether or not data should
                          be "sphered" (i.e., each dimension should be normalized such that
                          its mean and variance are zero and one, respectively.)  This
                          sphering normalization would be performed after all training
                          samples had been received but before the actual SVM model was
                          built.  The dimension-specific normalization constants would then
                          be applied to all future incoming vectors prior to submitting them
                          to the SVM library for inference."""),
          NodeSpecItem(name="mode", type="string", access="gs",
                       description="""We operate in two modes: 'classification' - the SVM is
                          used for top-level classification; 'feedback' - the SVM is a "temporary"
                          classified whose real purpose is simply to identify the most useful
                          features so that this information can be fed back to the feature
                          selection stage for pruning purposes.  The only difference is the
                          choice of kernel; for 'classification' mode, then kernel is selected
                          based on the value of the parameter 'kernelType'.  But in 'feedback'
                          mode, the kernel is always of type 'linear' regardless of the value
                          of 'kernelType'."""),
          NodeSpecItem(name="deterministic", type="bool", constraints="bool", access="cgs", value = False,
                      description="""Set true to seed random number generator so that the SVM
                          picks the same hyperplanes each time during learning. Only has an
                          effect when finishLearning is called."""),
          NodeSpecItem(name='numSVDSamples', type='PyObject', access='gcs', value=None,
                       description="""If not None, carries out SVD transformation after
                          that many samples have been seen. Otherwise, performs SVD when
                          switching to inference."""),
          NodeSpecItem(name='numSVDDims', type='PyObject', access='gc', value=None,
                       description="""Number of dimensions to keep after SVD.
                          If set to 'adaptive' the number is chosen automatically."""),
          NodeSpecItem(name='fractionOfMax', type='PyObject', access='gc', value=None,
                       description="""The smallest singular value which is retained
                          as a fraction of the largest singular value. This is used
                          only if numSVDDims=='adaptive'."""),
          NodeSpecItem(name='useAuxiliary', type='PyObject', access='gc', value=None,
                       description="""Whether or not the classifier should use auxiliary
                          input data."""),
          NodeSpecItem(name='justUseAuxiliary', type='PyObject', access='gc', value=None,
                       description="""Whether or not the classifier should ONLY use the auxiliary
                          input data."""),
          NodeSpecItem(name='discardProblem', type='bool', access='cgs',
                       description="""Whether to discard extra SVM data to save space. Set to False
                          to preserve the data for the PCA visualizer."""),
          NodeSpecItem(name='nta_cpp_svm_seed', type='int', access='cgs', value=-1,
                       description="""Seed for the C++ RNG"""),
          NodeSpecItem(name='keepSamples', type='bool', access='cgs', value=False),
          NodeSpecItem(name='calculateDistances', type='bool', access='cgs', value=False),
          NodeSpecItem(name='nta_monitorMemory', type='bool', access='cgs',
                       description="""Whether to spawn an internal thread that monitors the
                          amount of memory being consumed during processing and reports
                          statistics""", value=False),
          NodeSpecItem(name="numTrainingSamples", type="int", access="g", value = 0,
                       description="""Returns the current number of training samples."""),
          ]
      )
    return out + parent


  def catIndexToId(self, catIndex):
    """
    Map category indices (internal) to category IDs (external).
    """
    return self._catIdMap[catIndex]


  def _getHyperplanes(self):
    """
    Return a numpy array containing the complete set of hyperplanes used
    by the (trained) SVM classifier.  In general, there will be N(N-1)/2
    hyperplanes, where N is the number of categories.  Each hyperplane
    will be returned as a row within the numpy array; each row will be
    of dimensionality equal to the number of "features" presented to the
    SVM during training and inference.
    """
    hyperplanes = self._svm.get_model().get_hyperplanes()
    return hyperplanes


  def simulateTrainingSample(self, inputWidth=None, category=None, partitionId=None):
    """
    Debugging/profiling utility method to allow tools to
    simulate the presentation of training sample.
    """

    if inputWidth is None:
      inputWidth = self._inputWidth
    if category is None:
      category = 0

    if self._samples is None:
      self._samples = numpy.zeros((0, inputWidth), dtype=RealNumpyDType)
      assert self._labels is None
      self._labels = []
      assert self._partitionIds is None
      self._partitionIds = []

    # Add the sample vector and category lable
    sample = numpy.random.random((1, inputWidth))
    self._samples = numpy.concatenate((self._samples, sample), axis=0)
    self._labels += [category]
    if partitionId is not None:
      self._partitionIds.append(partitionId)


  def getParameter(self, parameterName, nodeSet=""):
    """
    Get the value of a parameter.

    Note: this method may be overridden by derived classes, but if so, then
    the derived class should invoke this base method if 'parameterName'
    is unknown to the derived class.

    @param parameterName -- the name of the parameter to retrieve, as defined
            by the Node Spec.
    """
    if parameterName == "hyperplanes":
      return self._getHyperplanes()
    elif parameterName in ("numCategoriesSeen", "categoryCount"):
      if self._inferenceMode:
        # Use the catIdMap, which is created during finishLearning
        return len(self._catIdMap)
      elif self._labels:
        # Use the labels, which are not thrown away until finishLearning
        # Don't count -1, if it appears
        return len(set(self._labels)) - int(-1 in self._labels)
      else:
        return 0
    elif parameterName == "numSVDSamples":
      return self._numSVDSamples
    elif parameterName == "numSVDDims":
      return self._numSVDDims
    elif parameterName == "fractionOfMax":
      return self._fractionOfMax
    elif parameterName == 'useAuxiliary':
      return self._useAuxiliary
    elif parameterName == "numTrainingSamples":
      return self.getNumTrainingSamples()
    else:
      return PyNode.getParameter(self, parameterName, nodeSet)

  def setParameter(self, parameterName, parameterValue, nodeSet=""):
    """Set the value of a parameter."""

    if parameterName == "numSVDSamples":
      self._numSVDSamples = parameterValue

    else:
      PyNode.setParameter(self, parameterName, parameterValue, nodeSet)


  def getNumTrainingSamples(self):
    if self._samples is None:
      numSamples = 0
    else:
      numSamples = self._samples.shape[0]
    return numSamples


  def _getActiveOutputCount(self):
    if self._catIdMap is not None:
      # Use the catIdMap, which is created during finishLearning
      return max(self._catIdMap)+1
    elif self._labels is not None:
      # Use the labels, which are not thrown away until finishLearning
      return max(self._labels)+1
    else:
      return 0

  activeOutputCount = property(fget=_getActiveOutputCount)

  # Support for using non-training data for (C, Gamma)
  # parameter optimization search.
  def _setAutoTuningData(self, value):
    self._autoTuningData = value
    if value and self._numSVDDims is not None and self._vt is None:
      # Run SVD now, so we can save space by projecting the autotuning samples
      self.computeSVD()
  def _getAutoTuningData(self):
    return self._autoTuningData
  autoTuningData = property(_getAutoTuningData, _setAutoTuningData)


  def __setstate__(self, state):
    """
    Set the state of ourself from a serialized state.
    """

    # Register a global variable for scanning or other tomfoolery
    PyNodeModule.nodes = getattr(PyNodeModule, 'nodes', []) + [self]

    self.__dict__.update(state)
    self._firstComputeCall = True

    # Backward compatibility to be able to load
    # saved networks in which these attributes were
    # not defined.
    missing = dict(_vt=None,
                   deterministic=False,
                   discardProblem=True,
                   cpp_svm_seed=-1,
                   keepSamples=False,
                   calculateDistances=False,
                   _partitionIds=[],
                   C=self.minC,
                   gamma=self.minGamma,
                   _autoTuningData=False,
                   _autoTuneSamples=None,
                   _autoTuneLabels=None,
                   _autoTunePartitionIds=None,
                   distances=None,
                   _upcomingPartitionIds=None,
                   _monitorMemory=False,
                   _distanceCount=0,
                   _adaptiveSVDDims=False,
                   _fractionOfMax=None,
                   _useAuxiliary=False,
                   _inputVector=None,
                   _auxInputLen=None,
                   _scanInfo=None,
                   _scanResults=None)

    d = self.__dict__
    for k in missing:
      d[k] = d.get(k, missing[k])

    self._initRandom()


  def __getstate__(self):
    """
    Return serializable state.  This function will return a version of the
    __dict__ with all "ephemeral" members stripped out.  "Ephemeral" members
    are defined as those that do not need to be (nor should be) stored
    in any kind of persistent file (e.g., NuPIC network XML file.)
    """
    state = self.__dict__.copy()
    for ephemeralMemberName in [x for x in self._getEphemeralMembers() if x in state]:
      del state[ephemeralMemberName]
    return state


  def _getEphemeralMembers(self):
    """
    Returns list of all ephemeral class members.
    """
    return [
        '_rng',
        '_autoTuneSamples',
        '_autoTuneLabels',
        '_autoTunePartitionIds',
        ]


  def _initDataStructures(self, inputWidth):
    """
    Initialize internal data structures.
    """
    self._inputWidth = inputWidth

    self._catIdMap = None

    self._initSvm()

  def _initSvm(self, n_dims=None):
    """Initialize SVM Engine: Use the SWIG bindings to initialize
    an instance of an SVM Classifier engine."""

    # Select kernel type. If we are in 'classification' mode,
    # then this choice is simply the user-specified parameter
    # 'kernelType'.  If we are in 'feedback' mode however, we
    # must use a 'linear' kernel.
    #   0 = linear
    #   1 = rbf
    if self._mode == 'feedback' or self._kernelType == 'linear':
      kernelType = 0
    elif self._kernelType == 'rbf':
      kernelType = 1

    if n_dims is None:
      n_dims = self._inputWidth

    if self.useSparseSvm:
      self._svm = svm_01(kernelType,
                         n_dims=n_dims,
                         threshold=self.inputThresh,
                         probability=self.useProbabilisticSvm,
                         seed=self.cpp_svm_seed)
    else:
      self._svm = svm_dense(kernelType,
                            n_dims=n_dims,
                            probability=self.useProbabilisticSvm,
                            seed=self.cpp_svm_seed)


  def _learn(self, inputVector, trueCatIndex, partitionId=None):
    """
    Store current input vector and associated category index.
    """
    assert self._svm is not None
    # If we are sphering, then we can't provide the data to the SVM
    # library until we have computed per-dimension normalization constants.
    # So instead, we'll just store each training sample.
    # Note: now we always do this, because the labels are used to build the
    # category mapping later
    self._storeSample(trueCatIndex, inputVector, partitionId)


  def _storeSample(self, trueCatIndex, inputVector, partitionId):
    """
    Store a training sample and associated category label
    """

    # If the incoming samples are not training data, but instead
    # are testing data to be used only for parameter optimization,
    # then store them separately.
    if self._autoTuningData:
      # If this is the first sample, then allocate a numpy array
      # of the appropriate size in which to store all samples.
      if self._autoTuneSamples is None:
        self._autoTuneSamples = numpy.zeros((0, inputVector.shape[0]),
                                            dtype=RealNumpyDType)
        assert self._autoTuneLabels is None
        self._autoTuneLabels = []
        assert self._autoTunePartitionIds is None
        self._autoTunePartitionIds = []

      # Add the sample vector and category lable
      self._autoTuneSamples = numpy.concatenate((self._autoTuneSamples,
                                                 numpy.atleast_2d(inputVector)), axis=0)
      self._autoTuneLabels += [trueCatIndex]
      if partitionId is not None:
        self._autoTunePartitionIds.append(partitionId)

    # Normal mode (incoming samples are training data)
    else:
      # If this is the first sample, then allocate a numpy array
      # of the appropriate size in which to store all samples.
      if self._samples is None:
        self._samples = numpy.zeros((0, self._inputWidth), dtype=RealNumpyDType)
        assert self._labels is None
        self._labels = []
        assert self._partitionIds is None
        self._partitionIds = []

      # Add the sample vector and category lable
      self._samples = numpy.concatenate((self._samples, numpy.atleast_2d(inputVector)), axis=0)
      self._labels += [trueCatIndex]
      if partitionId is not None:
        self._partitionIds.append(partitionId)


  def _infer(self, sample):
    """
    Consult SVM to classify input vector.
    """
    if self.calculateDistances:
      # Calculate distances in the original input space (pre-sphering, post-PCA)
      # if specified, effectively being a KNN
      self._calculateAndStoreDistance(sample)

    # If we are sphering, then apply normalization
    if self.doSphering:
      sample = (sample + self._normOffset) * self._normScale

    numCats = self._svm.get_model().n_class()

    if self.useProbabilisticSvm:
      belief = numpy.zeros(numCats, dtype=RealNumpyDType)
      prediction = int(self._svm.predict_probability(sample, belief))
    else:
      prediction = int(self._svm.predict(sample))
      belief = numpy.zeros(numCats, dtype=RealNumpyDType)
      belief[prediction] = 1.0
    return belief


  def _finishSphering(self):
    """
    Compute normalization constants for each feature dimension
    based on the collected training samples.  Then normalize our
    training samples using these constants (so that each input
    dimension has mean and variance of zero and one, respectively.)
    Then feed these "sphered" training samples into the underlying
    SVM model.
    """
    # If we are sphering our data, we need to compute the
    # per-dimension normalization constants
    # First normalize the means (to zero)
    self._normOffset = self._samples.mean(axis=0) * -1.0
    # Don't modify _samples in-place; it may be saved and used later
    samples = self._samples + self._normOffset
    # Now normalize the variances (to one).  However, we need to be
    # careful because the variance could conceivably be zero for one
    # or more dimensions.
    variance = samples.var(axis=0)
    variance[numpy.where(variance == 0.0)] = 1.0
    self._normScale  = 1.0 / numpy.sqrt(variance)
    samples *= self._normScale

    return samples


  def _finishLearning(self):
    """
    Use the C++ implementation to build an SVM Model.
    """
    self._beginMemoryMonitoring("_finishLearning")

    print 'Finish learning'

    # Set progress to 0
    PyNodeModule.finishLearningProgress = 0.0

    if self._upcomingPartitionIds:
      raise Exception("Switching to inference, but upcomingPartitionIds is "
                      "not empty: %s" % self._upcomingPartitionIds)

    # Compute the mapping of ImageSensor IDs to internal indices
    # Only insert ImageSensor IDs if they were seen during training
    # self._catIdMap = sorted(list(set(self._labels)))
    # the above causes bugs!!!
    self._catIdMap = []
    for label in self._labels:
      if label not in self._catIdMap and label != -1:
        self._catIdMap.append(label)
    if self._autoTuneLabels:
      # Use the autotune labels too; there could be categories never seen
      # during training that showed up during autotuning
      for label in self._autoTuneLabels:
        if label not in self._catIdMap and label != -1:
          self._catIdMap.append(label)

    # Do sphering if specified
    if self.doSphering:
      samples = self._finishSphering()

    # Run SVD if necessary
    if self._numSVDDims is not None and self._vt is None:
      self._beginMemoryMonitoring("computeSVD")
      self.computeSVD()
      self._endMemoryMonitoring("computeSVD")

    if not self.doSphering:
      samples = self._samples

    # Feed each sample into the SVM library
    for sampleIndex, label in enumerate(self._labels):
      # Ignore samples whose label is -1, which can occur if
      #  changeCategoriesOfPartitionIds was called
      # Use the catIdMap to give the SVM a continuous range of indices
      if label != -1:
        self._svm.add_sample(self._catIdMap.index(label), samples[sampleIndex])

    # Initialize the random number generator for LHS
    #self._initRandom()

    if self.numRecursions > 0 and (self.minC < self.maxC or self.minGamma < self.maxGamma):
      # Do recursive Latin Hypercube sampling to obtain best C, gamma
      paramRange = (self.minC, self.maxC, self.minGamma, self.maxGamma)
      self._beginMemoryMonitoring("_doRecursion")
      (svmParams, svmAccuracy, svmModel) = self._doRecursion(None, None, paramRange)
      self._endMemoryMonitoring("_doRecursion")
      if svmParams is None:
        # User canceled
        self._learningMode = False  # To avoid exception
        self._inferenceMode = True
        return
      print 'optimal params', svmParams
      # Set self.C and self.gamma to the optimal values
      self.C, self.gamma = svmParams
    else:
      # Use the specified values of C and gamma
      svmParams = (self.C, self.gamma)

    # Build final model
    self._beginMemoryMonitoring("_buildSVM")
    self._buildSVM(svmParams[0], svmParams[1])
    self._endMemoryMonitoring("_buildSVM")
    # Save parameters used to build the model
    self._svmParams = svmParams

    if self.discardProblem:
      # Discard the samples and labels
      self._svm.discard_problem()

    # Uncomment this to print out the chosen C and Gamma values
    #print "SVM node (C,Gamma) = ", self._svmParams

    # We can now throw away all our stored training samples; this way they
    # won't get pickled into the final 'inference stage' node.
    if not self.keepSamples:
      self._samples = None
      self._labels  = None

    # Set learning mode off and turn on inference mode
    self._learningMode = False
    self._inferenceMode = True

    # Clear data structures that should begin fresh for testing
    self.distances = None
    self._distanceCount = None

    # Update progress
    PyNodeModule.finishLearningProgress = 1.0

    self._endMemoryMonitoring("_finishLearning")


  def _doRecursion(self, samples, validationSets, paramRange, recursionIndex=0, bestResults=None):
    """
    Perform recursive Latin Hypercube sampling
    """
    # Construct set of sample points
    minC, maxC, minGamma, maxGamma = paramRange
    rangeC = maxC - minC
    rangeGamma = maxGamma - minGamma

    if self.numSamplesPerRecursion == 1:
      intervalC = 0.0
      intervalGamma = 0.0
    else:
      intervalC = rangeC / float(self.numSamplesPerRecursion - 1)
      intervalGamma = rangeGamma / float(self.numSamplesPerRecursion - 1)

    gammaIndices = range(self.numSamplesPerRecursion)
    self._rng.shuffle(gammaIndices)

    sampleIndices = [(float(x), float(gammaIndices[x])) for x in range(self.numSamplesPerRecursion)]

    samplePoints = [(minC + intervalC * x[0], minGamma + intervalGamma * x[1]) for x in sampleIndices]

    bestAccuracy = -1.0
    bestSamplePoint = None
    for i, samplePoint in enumerate(samplePoints):

      # For progress updates
      progressStart = PyNodeModule.finishLearningProgress
      progressEnd = progressStart \
                  + 1 / float(self.numSamplesPerRecursion * self.numRecursions)
      # Make sure progress does not reach 1.0 before the end of finishLearning
      progressEnd = min(0.99, progressEnd)

      logging.debug("SVM recursion %d, sample %d: C is %.2f, gamma is %.2f"
                    % (recursionIndex, i, samplePoint[0], samplePoint[1]))
      accuracy, svmModel = self._validateSvm(samplePoint[0], samplePoint[1],
                                             progressStart, progressEnd)
      if accuracy is None:
        # User canceled
        return (None, None, None)

      logging.debug("Accuracy: %.2f" % (accuracy * 100))
      if accuracy > bestAccuracy:
        bestAccuracy = accuracy
        bestSamplePoint = samplePoint
        bestModel = svmModel

    # Check if the previous recursion actually produced better results (by chance)
    if bestResults is not None and bestResults[1] > bestAccuracy:
      bestSamplePoint, bestAccuracy, bestModel = bestResults

    # Have we completed our LHS recursions?
    recursionIndex += 1
    if recursionIndex == self.numRecursions:
      return (bestSamplePoint, bestAccuracy, bestModel)

    # Perform another round of LHS
    else:
      # Contract down the sampling range
      halfRangeC = 0.5 * self.contractionFactor * rangeC
      halfRangeGamma = 0.5 * self.contractionFactor * rangeGamma

      newParamRange = (bestSamplePoint[0] - halfRangeC,
                       bestSamplePoint[0] + halfRangeC,
                       bestSamplePoint[1] - halfRangeGamma,
                       bestSamplePoint[1] + halfRangeGamma)

      return self._doRecursion(samples, validationSets, newParamRange, recursionIndex,
                              (bestSamplePoint, bestAccuracy, svmModel))


  def _autoTuneTest(self, svm):
    if self.useProbabilisticSvm:
      numCats = svm.get_model().n_class()
      belief = numpy.zeros(numCats, dtype=RealNumpyDType)
    numErrors = 0
    numTestingSamples = self._autoTuneSamples.shape[0]
    for k in xrange(numTestingSamples):
      sample = self._autoTuneSamples[k]
      # Note: we currently do not support sphering and PCA at
      # the same time.
      assert not self.doSphering or not self._vt
      # If we are sphering, then apply normalization
      if self.doSphering:
        sample = (sample + self._normOffset) * self._normScale
      # Vector is already projected onto PCA basis (if present)
      # Present the sample
      if self.useProbabilisticSvm:
        prediction = int(svm.predict_probability(sample, belief))
      else:
        prediction = int(svm.predict(sample))
      if prediction != self._catIdMap.index(self._autoTuneLabels[k]):
        numErrors += 1
    return float(numTestingSamples - numErrors) / float(numTestingSamples)


  def _validateSvm(self, C, gamma, progressStart, progressEnd):
    """
    Perform cross-validation to measure the recognition accuracy of an SVM.
    """
    # @todo Problem - the cross_validate() API doesn't accept C, gamma parameters!
    #accuracy = self._svm.cross_validate(C, gamma, nFold=self.numCrossValidations)
    #accuracy = self._svm.cross_validate(nFold=self.numCrossValidations)

    valueC = math.pow(10.0, C)
    valueGamma = math.pow(10.0, gamma)

    # Validate against test samples
    if self._autoTuneSamples is not None:
      assert self._autoTuneLabels is not None
      assert self._autoTunePartitionIds is not None

      if False:
          # Choose number of dimensions
          if self._numSVDDims:
            numDims = self._numSVDDims
          else:
            numDims = self._inputWidth
          # Choosing kernel type
          if self._mode == 'feedback' or self._kernelType == 'linear':
            kernelType = 0
          elif self._kernelType == 'rbf':
            kernelType = 1
          # Create a new svm
          if self.useSparseSvm:
            svm = svm_01(kernelType,
                            n_dims=numDims,
                            threshold=self.inputThresh,
                            probability=self.useProbabilisticSvm,
                            seed=self.cpp_svm_seed)
          else:
            svm = svm_dense(kernelType,
                            n_dims=numDims,
                            probability=self.useProbabilisticSvm,
                            seed=self.cpp_svm_seed)
          # Feed each sample into the SVM library
          for k, label in enumerate(self._labels):
            # Ignore samples whose label is -1, which can occur if
            #  changeCategoriesOfPartitionIds was called
            # Use the catIdMap to give the SVM a continuous range of indices
            sample = self._samples[k]
            if label != -1:
              svm.add_sample(self._catIdMap.index(label), sample)
      else:
        svm = self._svm

      # Train using all the training samples with this
      # (C, Gamma) sample point
      svm.trainReleaseGIL(gamma=valueGamma, C=valueC, eps=self.convEpsilon)

      # Check and update progress
      if PyNodeModule.finishLearningProgress == -1:
        return (None, None)
      PyNodeModule.finishLearningProgress += (progressStart - progressEnd)/2

      # Test against all testing samples
      accuracy = self._autoTuneTest(svm)

      # Check and update progress
      if PyNodeModule.finishLearningProgress == -1:
        return (None, None)
      PyNodeModule.finishLearningProgress = progressEnd

      #print "(%.3f, %.3f) ==> %.2f%%" % (C, gamma, 100.0 * accuracy)

    # Use cross-validation to validate against training samples
    else:
      accuracy = self._svm.cross_validate(n_fold=self.numCrossValidations,
                                                gamma=valueGamma,
                                                C=valueC,
                                                eps=self.convEpsilon)
    return (accuracy, None)


  def _buildSVM(self, C, gamma):
    """
    Train an SVM model.

    @param C -- the value of parameter C to use.
    @param gamma -- the value of parameter Gamma to use.
    @param trainLabels -- the category labels associated with the
            training data.
    @param trainSamples -- the input sample vectors associated with
            the training data.
    """
    valueC = math.pow(10.0, C)
    valueGamma = math.pow(10.0, gamma)
    self._svm.trainReleaseGIL(gamma=valueGamma, C=valueC, eps=self.convEpsilon)


  def computeSVD(self, numSVDSamples=None, finalize=True):
    print 'Computing SVD'

    # Samples are in self._samples, not in the SVM yet
    if numSVDSamples is None:
      numSVDSamples = self._samples.shape[0]

    if self._useAuxiliary and not self._justUseAuxiliary:
      self._mean = numpy.mean(self._samples[:,:self._samples.shape[1]-self._auxInputLen], axis=0)
      self._samples[:,:self._samples.shape[1]-self._auxInputLen] -= self._mean
      # Remove the auxiliary data prior to computing the SVD
      u, self._s, self._vt = numpy.linalg.svd(self._samples[:numSVDSamples,:self._samples.shape[1]-self._auxInputLen])
    else:
      self._mean = numpy.mean(self._samples, axis=0)
      self._samples -= self._mean
      u, self._s, self._vt = numpy.linalg.svd(self._samples[:numSVDSamples,:])


    if finalize:
      self.finalizeSVD()
    return self._s


  def getAdaptiveSVDDims(self, singularValues, fractionOfMax=0.001):
    v = singularValues/singularValues[0]
    idx = numpy.where(v<fractionOfMax)[0]
    if len(idx):
      print "Number of PCA dimensions chosen: ", idx[0], "out of ", len(v)
      return idx[0]
    else:
      print "Number of PCA dimensions chosen: ", len(v)-1, "out of ", len(v)
      return len(v)-1


  def finalizeSVD(self, numSVDDims=None):
    print 'Finalizing SVD'


    if numSVDDims is not None:
      self._numSVDDims = numSVDDims

    if self._numSVDDims=='adaptive':
      if self._fractionOfMax is not None:
        self._numSVDDims = self.getAdaptiveSVDDims(self._s, self._fractionOfMax)
      else:
        self._numSVDDims = self.getAdaptiveSVDDims(self._s)

    if self._vt.shape[0] < self._numSVDDims:
      print "******************************************************************************"
      print "Warning: The requested number of PCA dimensions is more than the number of pattern dimensions."
      print "Setting numSVDDims = ", self._vt.shape[0]
      print "******************************************************************************"
      self._numSVDDims = self._vt.shape[0]

    self._vt = self._vt[:self._numSVDDims]
    if self._useAuxiliary and not self._justUseAuxiliary:
      self._initSvm(n_dims=self._numSVDDims + self._auxInputLen)
      # Project all the vectors (mean has already been subtracted from each one)
      auxSamples = self._samples[:, self._samples.shape[1]+1:]
      self._samples = self._samples[:, :self._samples.shape[1]-self._auxInputLen]
      self._samples = numpy.dot(self._samples, self._vt.T)
      self._samples = numpy.concatenate([numpy.atleast_2d(self._samples), numpy.atleast_2d(auxSamples)], axis=1)
    else:
      self._initSvm(n_dims=self._numSVDDims)
      self._samples = numpy.dot(self._samples, self._vt.T)


  def getAllDistances(self):
    """Return all the prototype distances from all computes available."""

    if self.distances is None:
      return None
    return self.distances[:self._distanceCount, :]


  def getLatestDistances(self):
    """Get the distances to all training samples (pre-SVM, post-PCA)."""

    if self._inputVector is None:
      return
    if self._samples is None:
      raise Exception("No samples stored")

    return self._calculateDistances(self._inputVector)


  def getCategoryList(self):
    """
    Public API for returning the category list
    """

    return self._labels


  def _calculateDistances(self, inputVector):
    """Calculate distances in the original input space (pre-SVM, post-PCA)."""

    # Calculate the distances from this input to all prototypes
    # TODO do this after sphering instead?
    # TODO custom distance norm?
    dist = numpy.power(numpy.abs(self._samples - inputVector), 2)
    dist = dist.sum(1)
    dist = numpy.power(dist, 0.5)
    # Ignore samples with a category of -1
    dist[numpy.array(self._labels) == -1] = numpy.inf

    return dist


  def _calculateAndStoreDistances(self, inputVector):
    """Calculate distances in the original input space (pre-SVM, post-PCA)."""

    dist = self._calculateDistances(inputVector)

    # Keep all distances in an array
    if self.distances is None:
      self.distances = numpy.zeros((1, dist.shape[0]), dist.dtype)
      self.distances[0,:] = dist
      self._distanceCount = 1
    else:
      if self._distanceCount == self.distances.shape[0]:
        # Double the size of the array
        newDistances = numpy.zeros((self.distances.shape[0] * 2,
                                    self.distances.shape[1]),
                                   self.distances.dtype)
        newDistances[:self.distances.shape[0],:] = self.distances
        self.distances = newDistances
      # Store the new distances
      self.distances[self._distanceCount,:] = dist
      self._distanceCount += 1


  def setUpcomingPartitionIds(self, partitionIds):
    """
    Set the queue of upcoming partition ids. This can be used instead of the
    partitionId input. Checks that no partition ids currently exist (which
    could indicate a bug).
    """

    if self._upcomingPartitionIds:
      raise Exception("PartitionIds already exist: %s"
                      % str(self._upcomingPartitionIds))

    if not partitionIds:
      return

    if not hasattr(partitionIds, '__iter__'):
      partitionIds = [partitionIds]
    else:
      partitionIds = list(partitionIds)
    self._upcomingPartitionIds = partitionIds


  def remapCategories(self, mapping):
    """
    Change the existing category labels.

    mapping -- List of new category indices. For example, mapping=[2,0,1]
      would change all vectors of category 0 to be category 2, category 1 to 0,
      and category 2 to 1.
    """

    if not self._labels:
      return

    if not hasattr(mapping, '__len__'):
      mapping = [mapping]  # Cannot send singleton lists through session

    labels = numpy.array(self._labels, dtype=numpy.int)
    newLabels = numpy.zeros(labels.shape[0], dtype=numpy.int)
    newLabels.fill(-1)
    for i in xrange(len(mapping)):
      newLabels[labels==i] = mapping[i]
    self._labels = list(newLabels)


  def changePartitionId(self, oldPartitionId, newPartitionId):
    """
    Change all instances of oldPartitionId to newPartitionId.
    """

    if not self._partitionIds:
      # No learning has occurred yet
      return

    self._partitionIds = numpy.array(self._partitionIds)
    self._partitionIds[self._partitionIds == oldPartitionId] = newPartitionId
    self._partitionIds = list(self._partitionIds)


  def changeCategoriesOfPartitionIds(self, partitionIds, categoryIndices):
    """
    Change the category associated with all vectors with this partitionId(s).

    partitionIds -- Single id or list of ids.
    categoryIndices -- Single index or list of indices. Can also be a single
      index when partitionIds is a list, in which case the same category will
      be used for all vectors with the specified id.
    """

    if not hasattr(partitionIds, '__iter__'):
      partitionIds = [partitionIds]
      categoryIndices = [categoryIndices]
    elif not hasattr(categoryIndices, '__iter__'):
      categoryIndices = [categoryIndices] * len(partitionIds)

    if not self._partitionIds:
      # No learning has occurred yet
      return

    # Convert partitionIds and labels to arrays
    self._partitionIds = numpy.array(self._partitionIds)
    self._labels = numpy.array(self._labels)

    for i in xrange(len(partitionIds)):
      partitionId = partitionIds[i]
      categoryIndex = categoryIndices[i]
      self._labels[self._partitionIds == partitionId] = categoryIndex

    # Convert partitionIds and labels back to lists
    self._partitionIds = list(self._partitionIds)
    self._labels = list(self._labels)


  def switchToLearning(self):
    """Force a switch back to learning mode (not normally supported)."""

    self._learningMode = True
    self._inferenceMode = False

    # Clear data structures that were creating in finishLearning or first
    # used there
    self._catIdMap = None
    self._svm = None
    self._svmParams = None
    self._initSvm()
    self.distances = None
    self._distanceCount = None



if __name__=='__main__':
  import os
  from nupic.network import CreateNode

  name = os.path.splitext(os.path.basename(__file__))[0]
  n = CreateNode('nupic.pynodes.extra.%(m)s.%(m)s' % {'m':name})
  n.nodeHelp()



# ----------------------------------------------------------------------
# Numenta Platform for Intelligent Computing (NuPIC)
# Copyright (C) 2013, Numenta, Inc.  Unless you have an agreement
# with Numenta, Inc., for a separate license for this software code, the
# following terms and conditions apply:
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero Public License for more details.
#
# You should have received a copy of the GNU Affero Public License
# along with this program.  If not, see http://www.gnu.org/licenses.
#
# http://numenta.org/licenses/
# ----------------------------------------------------------------------

import numpy

def similar(a, b, eps=0.001):
  return (numpy.abs(a - b) < eps).all()


def lscsum(lx, epsilon=None):
  """
  Accepts log-values as input, exponentiates them, computes the sum,
  then converts the sum back to log-space and returns the result.
  Handles underflow by rescaling so that the largest values is exactly 1.0.
  """
  lx = numpy.asarray(lx)
  base = lx.max()

  # If the input is the log of 0's, catch this condition before we generate
  #  an exception, and return the log(0)
  if numpy.isinf(base):
    return base

  # If the user specified an epsilon and we are below it, return epsilon
  if (epsilon is not None) and (base < epsilon):
    return epsilon

  x = numpy.exp(lx - base)
  ssum = x.sum()

  result = numpy.log(ssum) + base
  # try:
  #   conventional = numpy.log(numpy.exp(lx).sum())
  #   if not similar(result, conventional):
  #     if numpy.isinf(conventional).any() and not numpy.isinf(result).any():
  #       # print "Scaled log sum avoided underflow or overflow."
  #       pass
  #     else:
  #       import sys
  #       print >>sys.stderr, "Warning: scaled log sum did not match."
  #       print >>sys.stderr, "Scaled log result:"
  #       print >>sys.stderr, result
  #       print >>sys.stderr, "Conventional result:"
  #       print >>sys.stderr, conventional
  # except FloatingPointError, e:
  #   # print "Scaled log sum avoided underflow or overflow."
  #   pass

  return result


def lscsum0(lx):
  """
  Accepts log-values as input, exponentiates them, sums down the rows
  (first dimension), then converts the sum back to log-space and returns the result.
  Handles underflow by rescaling so that the largest values is exactly 1.0.
  """
  # rows = lx.shape[0]
  # columns = numpy.prod(lx.shape[1:])
  # lx = lx.reshape(rows, columns)
  # bases = lx.max(1).reshape(rows, 1)
  # bases = lx.max(0).reshape((1,) + lx.shape[1:])
  lx = numpy.asarray(lx)
  bases = lx.max(0) # Don't need to reshape in the case of 0.
  x = numpy.exp(lx - bases)
  ssum = x.sum(0)

  result = numpy.log(ssum) + bases
  try:
    conventional = numpy.log(numpy.exp(lx).sum(0))

    if not similar(result, conventional):
      if numpy.isinf(conventional).any() and not numpy.isinf(result).any():
        # print "Scaled log sum down axis 0 avoided underflow or overflow."
        pass
      else:
        import sys
        print >>sys.stderr, "Warning: scaled log sum down axis 0 did not match."
        print >>sys.stderr, "Scaled log result:"
        print >>sys.stderr, result
        print >>sys.stderr, "Conventional result:"
        print >>sys.stderr, conventional
  except FloatingPointError, e:
    # print "Scaled log sum down axis 0 avoided underflow or overflow."
    pass


  return result


def normalize(lx):
  """
  Accepts log-values as input, exponentiates them,
  normalizes and returns the result.
  Handles underflow by rescaling so that the largest values is exactly 1.0.
  """
  lx = numpy.asarray(lx)
  base = lx.max()
  x = numpy.exp(lx - base)
  result = x / x.sum()

  conventional = (numpy.exp(lx) / numpy.exp(lx).sum())
  assert similar(result, conventional)

  return result


def nsum0(lx):
  """
  Accepts log-values as input, exponentiates them, sums down the rows
  (first dimension), normalizes and returns the result.
  Handles underflow by rescaling so that the largest values is exactly 1.0.
  """
  lx = numpy.asarray(lx)
  base = lx.max()
  x = numpy.exp(lx - base)
  ssum = x.sum(0)
  result = ssum / ssum.sum()

  conventional = (numpy.exp(lx).sum(0) / numpy.exp(lx).sum())
  assert similar(result, conventional)

  return result


def lnsum0(lx):
  """
  Accepts log-values as input, exponentiates them, sums down the rows
  (first dimension), normalizes, then converts the sum back to
  log-space and returns the result.
  Handles underflow by rescaling so that the largest values is exactly 1.0.
  """
  lx = numpy.asarray(lx)
  base = lx.max()
  x = numpy.exp(lx - base)
  ssum = x.sum(0)
  normalized = nsum0(lx)

  result = numpy.log(normalized)
  conventional = numpy.log(numpy.exp(lx).sum(0) / numpy.exp(lx).sum())
  assert similar(result, conventional)

  return result


def logSumExp(A, B, out=None):
  """ returns log(exp(A) + exp(B)). A and B are numpy arrays"""

  if out is None:
    out = numpy.zeros(A.shape)

  indicator1 = A >= B
  indicator2 = numpy.logical_not(indicator1)
  out[indicator1] = A[indicator1] + numpy.log1p(numpy.exp(B[indicator1]-A[indicator1]))
  out[indicator2]  = B[indicator2] + numpy.log1p(numpy.exp(A[indicator2]-B[indicator2]))

  return out

def logDiffExp(A, B, out=None):
  """ returns log(exp(A) - exp(B)). A and B are numpy arrays. values in A should be
  greater than or equal to corresponding values in B"""

  if out is None:
    out = numpy.zeros(A.shape)

  indicator1 = A >= B
  assert indicator1.all(), "Values in the first array should be greater than the values in the second"
  out[indicator1] = A[indicator1] + numpy.log(1 - numpy.exp(B[indicator1]-A[indicator1]))

  return out



# ----------------------------------------------------------------------
# Numenta Platform for Intelligent Computing (NuPIC)
# Copyright (C) 2014, Numenta, Inc.  Unless you have an agreement
# with Numenta, Inc., for a separate license for this software code, the
# following terms and conditions apply:
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero Public License for more details.
#
# You should have received a copy of the GNU Affero Public License
# along with this program.  If not, see http://www.gnu.org/licenses.
#
# http://numenta.org/licenses/
# ----------------------------------------------------------------------

"""
Utilities for generating and manipulating patterns, for use in
experimentation and tests.
"""

import numpy as np
from nupic.bindings.math import Random



class PatternMachine(object):
  """
  Base pattern machine class.
  """

  def __init__(self,
               n,
               w,
               num=100,
               seed=42):
    """
    @param n   (int)      Number of available bits in pattern
    @param w   (int/list) Number of on bits in pattern
                          If list, each pattern will have a `w` randomly
                          selected from the list.
    @param num (int)      Number of available patterns
    """
    # Save member variables
    self._n = n
    self._w = w
    self._num = num

    # Initialize member variables
    self._random = Random(seed)
    self._patterns = dict()

    self._generate()


  def get(self, number):
    """
    Return a pattern for a number.

    @param number (int) Number of pattern

    @return (set) Indices of on bits
    """
    if not number in self._patterns:
      raise IndexError("Invalid number")

    return self._patterns[number]


  def addNoise(self, bits, amount):
    """
    Add noise to pattern.

    @param bits   (set)   Indices of on bits
    @param amount (float) Probability of switching an on bit with a random bit

    @return (set) Indices of on bits in noisy pattern
    """
    newBits = set()

    for bit in bits:
      if self._random.getReal64() < amount:
        newBits.add(self._random.getUInt32(self._n))
      else:
        newBits.add(bit)

    return newBits


  def numbersForBit(self, bit):
    """
    Return the set of pattern numbers that match a bit.

    @param bit (int) Index of bit

    @return (set) Indices of numbers
    """
    if bit >= self._n:
      raise IndexError("Invalid bit")

    numbers = set()

    for index, pattern in self._patterns.iteritems():
      if bit in pattern:
        numbers.add(index)

    return numbers


  def numberMapForBits(self, bits):
    """
    Return a map from number to matching on bits,
    for all numbers that match a set of bits.

    @param bits (set) Indices of bits

    @return (dict) Mapping from number => on bits.
    """
    numberMap = dict()

    for bit in bits:
      numbers = self.numbersForBit(bit)

      for number in numbers:
        if not number in numberMap:
          numberMap[number] = set()

        numberMap[number].add(bit)

    return numberMap


  def prettyPrintPattern(self, bits, verbosity=1):
    """
    Pretty print a pattern.

    @param bits      (set) Indices of on bits
    @param verbosity (int) Verbosity level

    @return (string) Pretty-printed text
    """
    numberMap = self.numberMapForBits(bits)
    text = ""

    numberList = []
    numberItems = sorted(numberMap.iteritems(),
                         key=lambda (number, bits): len(bits),
                         reverse=True)

    for number, bits in numberItems:

      if verbosity > 2:
        strBits = [str(n) for n in bits]
        numberText = "{0} (bits: {1})".format(number, ",".join(strBits))
      elif verbosity > 1:
        numberText = "{0} ({1} bits)".format(number, len(bits))
      else:
        numberText = str(number)

      numberList.append(numberText)

    text += "[{0}]".format(", ".join(numberList))

    return text


  def _generate(self):
    """
    Generates set of random patterns.
    """
    candidates = np.array(range(self._n), np.uint32)
    for i in xrange(self._num):
      self._random.shuffle(candidates)
      pattern = candidates[0:self._getW()]
      self._patterns[i] = set(pattern)


  def _getW(self):
    """
    Gets a value of `w` for use in generating a pattern.
    """
    w = self._w

    if type(w) is list:
      return w[self._random.getUInt32(len(w))]
    else:
      return w



class ConsecutivePatternMachine(PatternMachine):
  """
  Pattern machine class that generates patterns with non-overlapping,
  consecutive on bits.
  """

  def _generate(self):
    """
    Generates set of consecutive patterns.
    """
    n = self._n
    w = self._w

    assert type(w) is int, "List for w not supported"

    for i in xrange(n / w):
      pattern = set(xrange(i * w, (i+1) * w))
      self._patterns[i] = pattern



# ----------------------------------------------------------------------
# Numenta Platform for Intelligent Computing (NuPIC)
# Copyright (C) 2013, Numenta, Inc.  Unless you have an agreement
# with Numenta, Inc., for a separate license for this software code, the
# following terms and conditions apply:
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero Public License for more details.
#
# You should have received a copy of the GNU Affero Public License
# along with this program.  If not, see http://www.gnu.org/licenses.
#
# http://numenta.org/licenses/
# ----------------------------------------------------------------------

from nupic.encoders.base import Encoder
from nupic.encoders.scalar import ScalarEncoder
from nupic.encoders.adaptivescalar import AdaptiveScalarEncoder
from nupic.encoders.date import DateEncoder
from nupic.encoders.logenc import LogEncoder
from nupic.encoders.category import CategoryEncoder
from nupic.encoders.sdrcategory import SDRCategoryEncoder
from nupic.encoders.delta import DeltaEncoder
from nupic.encoders.scalarspace import ScalarSpaceEncoder
from nupic.encoders.pass_through_encoder import PassThroughEncoder
from nupic.encoders.sparse_pass_through_encoder import SparsePassThroughEncoder
from nupic.encoders.coordinate import CoordinateEncoder
from nupic.encoders.geospatial_coordinate import GeospatialCoordinateEncoder
# multiencoder must be imported last because it imports * from this module!
from nupic.encoders.utils import bitsToString
from nupic.encoders.random_distributed_scalar import RandomDistributedScalarEncoder



# Map class to Cap'n Proto schema union attribute
_CLASS_ATTR_MAP = {
  ScalarEncoder: "scalarEncoder",
  AdaptiveScalarEncoder: "adaptiveScalarEncoder",
  DateEncoder: "dateEncoder",
  LogEncoder: "logEncoder",
  CategoryEncoder: "categoryEncoder",
  CoordinateEncoder: "coordinateEncoder",
  SDRCategoryEncoder: "sdrCategoryEncoder",
  DeltaEncoder: "deltaEncoder",
  PassThroughEncoder: "passThroughEncoder",
  SparsePassThroughEncoder: "sparsePassThroughEncoder",
  RandomDistributedScalarEncoder: "randomDistributedScalarEncoder"
}

# Invert for fast lookup in MultiEncoder.read()
_ATTR_CLASS_MAP = {value:key for key, value in _CLASS_ATTR_MAP.items()}




class MultiEncoder(Encoder):
  """A MultiEncoder encodes a dictionary or object with
  multiple components. A MultiEncode contains a number
  of sub-encoders, each of which encodes a separate component."""
  # TODO expand this docstring to explain how the multiple encoders are combined


  def __init__(self, encoderDescriptions=None):
    self.width = 0
    self.encoders = []
    self.description = []
    self.name = ''
    if encoderDescriptions is not None:
      self.addMultipleEncoders(encoderDescriptions)


  def setFieldStats(self, fieldName, fieldStatistics ):
    for (name, encoder, offset) in self.encoders:
      encoder.setFieldStats(name, fieldStatistics)


  def addEncoder(self, name, encoder):
    self.encoders.append((name, encoder, self.width))
    for d in encoder.getDescription():
      self.description.append((d[0], d[1] + self.width))
    self.width += encoder.getWidth()

    self._flattenedEncoderList = None
    self._flattenedFieldTypeList = None


  def encodeIntoArray(self, obj, output):
    for name, encoder, offset in self.encoders:
        encoder.encodeIntoArray(self._getInputValue(obj, name), output[offset:])


  def getDescription(self):
    return self.description


  def getWidth(self):
    """Represents the sum of the widths of each fields encoding."""
    return self.width

  def setLearning(self,learningEnabled):
    encoders = self.getEncoderList()
    for encoder in encoders:
      encoder.setLearning(learningEnabled)
    return


  def encodeField(self, fieldName, value):
    for name, encoder, offset in self.encoders:
      if name == fieldName:
        return encoder.encode(value)


  def encodeEachField(self, inputRecord):
    encodings = []
    for name, encoder, offset in self.encoders:
      encodings.append(encoder.encode(getattr(inputRecord, name)))
    return encodings


  def addMultipleEncoders(self, fieldEncodings):
    """
    fieldEncodings -- a dict of dicts, mapping field names to the field params
                        dict.

    Each field params dict has the following keys
    1) data fieldname that matches the key ('fieldname')
    2) an encoder type ('type')
    3) and the encoder params (all other keys)

    For example,
    fieldEncodings={
        'dateTime': dict(fieldname='dateTime', type='DateEncoder',
                         timeOfDay=(5,5)),
        'attendeeCount': dict(fieldname='attendeeCount', type='ScalarEncoder',
                              name='attendeeCount', minval=0, maxval=250,
                              clipInput=True, w=5, resolution=10),
        'consumption': dict(fieldname='consumption',type='ScalarEncoder',
                            name='consumption', minval=0,maxval=110,
                            clipInput=True, w=5, resolution=5),
    }
    
    would yield a vector with a part encoded by the DateEncoder, 
    and to parts seperately taken care of by the ScalarEncoder with the specified parameters. 
    The three seperate encodings are then merged together to the final vector, in such a way that
    they are always at the same location within the vector.
    """

    # Sort the encoders so that they end up in a controlled order
    encoderList = sorted(fieldEncodings.items())
    for key, fieldParams in encoderList:
      if ':' not in key and fieldParams is not None:
        fieldParams = fieldParams.copy()
        fieldName   = fieldParams.pop('fieldname')
        encoderName = fieldParams.pop('type')
        try:
          self.addEncoder(fieldName, eval(encoderName)(**fieldParams))
        except TypeError, e:
          print ("#### Error in constructing %s encoder. Possibly missing "
                "some required constructor parameters. Parameters "
                "that were provided are: %s" %  (encoderName, fieldParams))
          raise


  @classmethod
  def read(cls, proto):
    encoder = object.__new__(cls)

    encoder.encoders = [None] * len(proto.encoders)

    encoder.width = 0

    for index, encoderProto in enumerate(proto.encoders):
      # Identify which attr is set in union
      encoderType = encoderProto.which()

      encoderDetails = getattr(encoderProto, encoderType)
      encoder.encoders[index] = (
        encoderProto.name,
        # Call class.read() where class is determined by _ATTR_CLASS_MAP
        _ATTR_CLASS_MAP.get(encoderType).read(encoderDetails),
        encoderProto.offset
      )

      encoder.width += encoder.encoders[index][1].getWidth()

    # Derive description from encoder list
    encoder.description = [(enc[1].name, enc[2]) for enc in encoder.encoders]
    encoder.name = proto.name

    return encoder


  def write(self, proto):

    proto.init("encoders", len(self.encoders))

    for index, (name, encoder, offset) in enumerate(self.encoders):
      encoderProto = proto.encoders[index]
      encoderType = _CLASS_ATTR_MAP.get(encoder.__class__)
      encoderProto.init(encoderType)
      encoderDetails = getattr(encoderProto, encoderType)
      encoder.write(encoderDetails)
      encoderProto.name = name
      encoderProto.offset = offset

    proto.name = self.name



# ----------------------------------------------------------------------
# Numenta Platform for Intelligent Computing (NuPIC)
# Copyright (C) 2013, Numenta, Inc.  Unless you have an agreement
# with Numenta, Inc., for a separate license for this software code, the
# following terms and conditions apply:
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero Public License for more details.
#
# You should have received a copy of the GNU Affero Public License
# along with this program.  If not, see http://www.gnu.org/licenses.
#
# http://numenta.org/licenses/
# ----------------------------------------------------------------------


# TODO for NUPIC 2 -- document the interface!
# TODO for NuPIC 2 -- should this move to inferenceanalysis?




def _calculateColumnsFromLine(line):
  if "," in line:
    splitLine = line.strip().split(",")
    n = len(splitLine)
    if n:
      if not splitLine[-1].strip():
        return n-1
      else:
        return n
    else:
      return 0
  else:
    # Too flexible.
    # return len([x for x in line.strip().split() if x != ","])
    return len(line.strip().split())

def _isComment(strippedLine):
  if strippedLine:
    return strippedLine.startswith("#")
  else:
    return True

def _calculateColumnsFromFile(f, format, rewind):
  # Calculate the number of columns.
  # We will put more trust in the second line that the first, in case the
  # first line includes header entries.
  if format not in [0, 2, 3]:
    raise RuntimeError("Supported formats are 0, 2, and 3.")

  if format == 0:
    line0 = f.readline()
    csplit = line0.split()
    if len(csplit) != 1:
      raise RuntimeError("Expected first line of data file to  "
                         "contain a single number of columns. "
                         " Found %d fields" % len(csplit))
    try:
      numColumns = int(csplit[0])
    except:
      raise RuntimeError("Expected first line of data file to "
                         "contain a single number of columns. Found '%s'" % csplit[0])
    if rewind:
      f.seek(0)

    return numColumns

  elif format == 2:
    numColumns = 0
    numLinesRead = 0
    for line in f:
      strippedLine = line.strip()
      if not _isComment(strippedLine):
        curColumns = _calculateColumnsFromLine(strippedLine)
        numLinesRead += 1
        if numColumns and (numColumns != curColumns):
          raise RuntimeError("Different lines have different "
            "numbers of columns.")
        else:
          numColumns = curColumns
        if numLinesRead > 1:
          break
    if rewind:
      f.seek(0)
    return numColumns

  # CSV file: we'll just check the first line
  elif format == 3:
    strippedLine = f.readline().strip()
    numColumns = calculateColumnsFromLine(strippedLine)
    if rewind:
      f.seek(0)
    return numColumns

def processCategoryFile(f, format, categoryColumn=None, categoryColumns=None, count=1):
  """Read the data out of the given category file, returning a tuple
  (categoryCount, listOfCategories)

  @param  f                A file-like object containing the category info.
  @param  format           The format of the category file.  TODO: describe.
  @param  categoryColumn   If non-None, this is the column number (zero-based)
                           where the category info starts in the file.  If
                           None, indicates that the file only contains category
                           information (same as passing 0, but allows some
                           extra sanity checking).
  @param  categoryColumns  Indicates how many categories are active per
                           timepoint (how many elements wide the category info
                           is).  If 0, we'll determine this from the file.  If
                           None (the default), means that the category info
                           is 1 element wide, and that the list we return
                           will just be a list of ints (rather than a list of
                           lists)
  @param  count            Determines the size of chunks that will be aggregated
                           into a single entry. The default is 1, so each entry
                           from the file will be represented in the result. If
                           count > 1 then 'count' categories (all identical) will
                           be collapsed into a single entry. This is helpful for
                           aggregating explorers like EyeMovements where multiple
                           presentaions are conceptually the same item.
  @return categoryCount    The number of categories (aka maxCat + 1)
  @return allCategories    A list of the categories read in, with one item per
                           time point.  If 'categoryColumns' is None, each item
                           will be an int.  Otherwise, each item will be a list
                           of ints. If count > 1 then the categories will be
                           aggregated, so that each chunk of 'count' categories
                           will result in only one entry (all categories in a chunk
                           must be identical)
  """
  calculatedCategoryColumns = _calculateColumnsFromFile(f, format=format,
    rewind=(format==2 or format==3))

  # If the user passed categoryColumns as None, we'll return a list of ints
  # directly; otherwise we'll return a list of lists...
  wantListOfInts = (categoryColumns is None)

  # Get arguments sanitized...
  if categoryColumns == 0:
    # User has told us to auto-calculate the # of categories / time point...

    # If categoryColumn is not 0 or None, that's an error...
    if categoryColumn:
      raise RuntimeError("You can't specify an offset for category data "
                         "if using automatic width.")

    categoryColumn = 0
    categoryColumns = calculatedCategoryColumns
  elif categoryColumns is None:
    # User has told us that there's just one category...

    if categoryColumn is None:
      if calculatedCategoryColumns != 1:
        raise RuntimeError("Category file must contain exactly one column.")
      categoryColumn = 0

    categoryColumns = 1
  else:
    # User specified exactly how big the category data is...

    if (categoryColumns + categoryColumn) > calculatedCategoryColumns:
      raise RuntimeError("Not enough categories in file")

  maxCategory = 0

  allCategories = []
  for line in f:
    strippedLine = line.strip()
    if not _isComment(strippedLine):
      if wantListOfInts:
        category = int(strippedLine.split()[categoryColumn])
        allCategories.append(category)
        maxCategory = max(maxCategory, category)
      else:
        categories = strippedLine.split()[categoryColumn:
                                          categoryColumn+categoryColumns]
        categories = map(int, categories)
        allCategories.append(categories)
        maxCategory = max(maxCategory, max(categories))

  categoryCount = maxCategory + 1

  # Aggregate categories
  result = []
  if count > 1:
    # Make sure there the number of categories can be aggregated
    # exactly by chunks of size 'count'
    assert len(allCategories) % count == 0
    start = 0
    for i in range(len(allCategories) / count):
      end = start + count
      # Make sure each chunk of size 'count' contains exactly one category
      assert (min(allCategories[start:end]) == max(allCategories[start:end]))
      # Add just one entry for each chunk
      result.append(allCategories[start])
      start = end
  else:
    result = allCategories

  return categoryCount, result



# ----------------------------------------------------------------------
# Numenta Platform for Intelligent Computing (NuPIC)
# Copyright (C) 2013, Numenta, Inc.  Unless you have an agreement
# with Numenta, Inc., for a separate license for this software code, the
# following terms and conditions apply:
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero Public License for more details.
#
# You should have received a copy of the GNU Affero Public License
# along with this program.  If not, see http://www.gnu.org/licenses.
#
# http://numenta.org/licenses/
# ----------------------------------------------------------------------

import os
import imp

from nupic.data.dictutils import rUpdate


# This file contains utility functions that are used
# internally by the prediction framework and may be imported
# by description files. Functions that are used only by
# the prediction framework should be in utils.py
#
# This file provides support for the following experiment description features:
#
# 1. Sub-experiment support
# 2. Lazy evaluators (e.g., DeferredDictLookup, applyValueGettersToContainer)


###############################################################################
# Sub-experiment support
###############################################################################


# Utility methods for description files  are organized as a base description
#  and an experiment based on that base description.
# The base description calls getConfig to get the configuration from the
# specific experiment, and the specific experiment calls importBaseDescription

# empty initial config allows base experiment to run by itself
_config = dict()

# Save the path to the current sub-experiment here during importBaseDescription()
subExpDir = None


# We will load the description file as a module, which allows us to
# use the debugger and see source code. But description files are frequently
# modified and we want to be able to easily reload them. To facilitate this,
# we reload with a unique module name ("pf_description%d") each time.
baseDescriptionImportCount = 0



def importBaseDescription(path, config):
  global baseDescriptionImportCount, _config, subExpDir
  if not os.path.isabs(path):
    # grab the path to the file doing the import
    import inspect
    callingFrame = inspect.stack()[1][0]
    callingFile = callingFrame.f_globals['__file__']
    subExpDir = os.path.dirname(callingFile)
    path = os.path.normpath(os.path.join(subExpDir, path))

  #print "Importing from: %s" % path

  # stash the config in a place where the loading module can find it.
  _config = config
  mod = imp.load_source("pf_base_description%d" % baseDescriptionImportCount,
                        path)
  # don't want to override __file__ in our caller
  mod.__base_file__ = mod.__file__
  del mod.__file__
  baseDescriptionImportCount += 1
  return mod



def updateConfigFromSubConfig(config):
  # Newer method just updates from sub-experiment
  # _config is the configuration provided by the sub-experiment
  global _config
  rUpdate(config, _config)
  _config = dict()



def getSubExpDir():
  global subExpDir
  return subExpDir





###############################################################################
# Lazy evaluators (DeferredDictLookup, applyValueGettersToContainer, and friends)
###############################################################################



class ValueGetterBase(object):
  """ Base class for "value getters" (e.g., class DictValueGetter) that are used
  to resolve values of sub-fields after the experiment's config dictionary (in
  description.py) is defined and possibly updated from a sub-experiment.

  This solves the problem of referencing the config dictionary's field from within
  the definition of the dictionary itself (before the dictionary's own defintion
  is complete).

  NOTE: its possible that the referenced value does not yet exist at the
        time of instantiation of a given value-getter future. It will be
        resolved when the base description.py calls
        applyValueGettersToContainer().

  NOTE: The constructor of the derived classes MUST call our constructor.
  NOTE: The derived classes MUST override handleGetValue(self).

  NOTE: may be used by base and sub-experiments to derive their own custom value
    getters; however, their use is applicapble only where permitted, as
    described in comments within descriptionTemplate.tpl. See class
    DictValueGetter for implementation example.
  """

  class __NoResult(object):
    """ A private class that we use as a special unique value to indicate that
    our result cache instance variable does not hold a valid result.
    """
    pass


  def __init__(self):
    #print("NOTE: ValueGetterBase INITIALIZING")
    self.__inLookup = False
    self.__cachedResult = self.__NoResult


  def __call__(self, topContainer):
    """ Resolves the referenced value.  If the result is already cached,
    returns it to caller. Otherwise, invokes the pure virtual method
    handleGetValue.  If handleGetValue() returns another value-getter, calls
    that value-getter to resolve the value.  This may result in a chain of calls
    that terminates once the value is fully resolved to a non-value-getter value.
    Upon return, the value is fully resolved and cached, so subsequent calls will
    always return the cached value reference.

    topContainer: The top-level container (dict, tuple, or list [sub-]instance)
                  within whose context the value-getter is applied.

    Returns:  The fully-resolved value that was referenced by the value-getter
              instance
    """

    #print("IN ValueGetterBase.__CAll__()")

    assert(not self.__inLookup)

    if self.__cachedResult is not self.__NoResult:
      return self.__cachedResult

    self.__cachedResult = self.handleGetValue(topContainer)

    if isinstance(self.__cachedResult, ValueGetterBase):
      valueGetter = self.__cachedResult
      self.__inLookup = True
      self.__cachedResult = valueGetter(topContainer)
      self.__inLookup = False

    # The value should be full resolved at this point
    assert(self.__cachedResult is not self.__NoResult)
    assert(not isinstance(self.__cachedResult, ValueGetterBase))

    return self.__cachedResult


  def handleGetValue(self, topContainer):
    """ A "pure virtual" method.  The derived class MUST override this method
    and return the referenced value.  The derived class is NOT responsible for
    fully resolving the reference'd value in the event the value resolves to
    another ValueGetterBase-based instance -- this is handled automatically
    within ValueGetterBase implementation.

    topContainer: The top-level container (dict, tuple, or list [sub-]instance)
                  within whose context the value-getter is applied.

    Returns:      The value referenced by this instance (which may be another
                  value-getter instance)
    """
    raise NotImplementedError("ERROR: ValueGetterBase is an abstract " + \
                              "class; base class MUST override handleGetValue()")



class DictValueGetter(ValueGetterBase):
  """
    Creates a "future" reference to a value within a top-level or a nested
    dictionary.  See also class DeferredDictLookup.
  """
  def __init__(self, referenceDict, *dictKeyChain):
    """
      referenceDict: Explicit reference dictionary that contains the field
                    corresonding to the first key name in dictKeyChain.  This may
                    be the result returned by the built-in globals() function,
                    when we desire to look up a dictionary value from a dictionary
                    referenced by a global variable within the calling module.
                    If None is passed for referenceDict, then the topContainer
                    parameter supplied to handleGetValue() will be used as the
                    reference dictionary instead (this allows the desired module
                    to designate the appropriate reference dictionary for the
                    value-getters when it calls applyValueGettersToContainer())

      dictKeyChain: One or more strings; the first string is a key (that will
                    eventually be defined) in the reference dictionary. If
                    additional strings are supplied, then the values
                    correspnding to prior key strings must be dictionaries, and
                    each additionl string references a sub-dictionary of the
                    former. The final string is the key of the field whose value
                    will be returned by handleGetValue().

      NOTE: Its possible that the referenced value does not yet exist at the
            time of instantiation of this class.  It will be resolved when the
            base description.py calls applyValueGettersToContainer().


    Example:
      config = dict(
        _dsEncoderFieldName2_N = 70,
        _dsEncoderFieldName2_W = 5,

        dsEncoderSchema = [
          dict(
            base=dict(
              fieldname='Name2', type='ScalarEncoder',
              name='Name2', minval=0, maxval=270, clipInput=True,
              n=DictValueGetter(None, '_dsEncoderFieldName2_N'),
              w=DictValueGetter(None, '_dsEncoderFieldName2_W')),
            ),
          ],
      )

      updateConfigFromSubConfig(config)
      applyValueGettersToContainer(config)
    """

    # First, invoke base constructor
    ValueGetterBase.__init__(self)

    assert(referenceDict is None or isinstance(referenceDict, dict))
    assert(len(dictKeyChain) >= 1)

    self.__referenceDict = referenceDict
    self.__dictKeyChain = dictKeyChain


  def handleGetValue(self, topContainer):
    """ This method overrides ValueGetterBase's "pure virtual" method.  It
    returns the referenced value.  The derived class is NOT responsible for
    fully resolving the reference'd value in the event the value resolves to
    another ValueGetterBase-based instance -- this is handled automatically
    within ValueGetterBase implementation.

    topContainer: The top-level container (dict, tuple, or list [sub-]instance)
                  within whose context the value-getter is applied.  If
                  self.__referenceDict is None, then topContainer will be used
                  as the reference dictionary for resolving our dictionary key
                  chain.

    Returns:      The value referenced by this instance (which may be another
                  value-getter instance)
    """
    value = self.__referenceDict if self.__referenceDict is not None else topContainer
    for key in self.__dictKeyChain:
      value = value[key]

    return value



class DeferredDictLookup(DictValueGetter):
  """
    Creates a "future" reference to a value within an implicit dictionary that
    will be passed to applyValueGettersToContainer() in the future (typically
    called by description.py after its config dictionary has been updated from
    the sub-experiment). The reference is relative to the dictionary that will
    be passed to applyValueGettersToContainer()
  """
  def __init__(self, *dictKeyChain):
    """
      dictKeyChain: One or more strings; the first string is a key (that will
                    eventually be defined) in the dictionary that will be passed
                    to applyValueGettersToContainer(). If additional strings are
                    supplied, then the values correspnding to prior key strings
                    must be dictionaries, and each additionl string references a
                    sub-dictionary of the former. The final string is the key of
                    the field whose value will be returned by this value-getter

      NOTE: its possible that the referenced value does not yet exist at the
            time of instantiation of this class.  It will be resolved when the
            base description.py calls applyValueGettersToContainer().


    Example:
      config = dict(
        _dsEncoderFieldName2_N = 70,
        _dsEncoderFieldName2_W = 5,

        dsEncoderSchema = [
          dict(
            base=dict(
              fieldname='Name2', type='ScalarEncoder',
              name='Name2', minval=0, maxval=270, clipInput=True,
              n=DeferredDictLookup('_dsEncoderFieldName2_N'),
              w=DeferredDictLookup('_dsEncoderFieldName2_W')),
            ),
          ],
      )

      updateConfigFromSubConfig(config)
      applyValueGettersToContainer(config)
    """

    # Invoke base (DictValueGetter constructor), passing None for referenceDict,
    # which will force it use the dictionary passed via
    # applyValueGettersToContainer(), instead.
    DictValueGetter.__init__(self, None, *dictKeyChain)



def applyValueGettersToContainer(container):
  """
  """
  _applyValueGettersImpl(container=container, currentObj=container,
                         recursionStack=[])



def _applyValueGettersImpl(container, currentObj, recursionStack):
  """
  """

  # Detect cycles
  if currentObj in recursionStack:
    return

  # Sanity-check of our cycle-detection logic
  assert(len(recursionStack) < 1000)

  # Push the current object on our cycle-detection stack
  recursionStack.append(currentObj)

  # Resolve value-getters within dictionaries, tuples and lists

  if isinstance(currentObj, dict):
    for (key, value) in currentObj.items():
      if isinstance(value, ValueGetterBase):
        currentObj[key] = value(container)

      _applyValueGettersImpl(container, currentObj[key], recursionStack)


  elif isinstance(currentObj, tuple) or isinstance(currentObj, list):
    for (i, value) in enumerate(currentObj):
      # NOTE: values within a tuple should never be value-getters, since
      #       the top-level elements within a tuple are immutable. However,
      #       if any nested sub-elements might be mutable
      if isinstance(value, ValueGetterBase):
        currentObj[i] = value(container)

      _applyValueGettersImpl(container, currentObj[i], recursionStack)

  else:
    pass

  recursionStack.pop()
  return



# ----------------------------------------------------------------------
# Numenta Platform for Intelligent Computing (NuPIC)
# Copyright (C) 2013, Numenta, Inc.  Unless you have an agreement
# with Numenta, Inc., for a separate license for this software code, the
# following terms and conditions apply:
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero Public License for more details.
#
# You should have received a copy of the GNU Affero Public License
# along with this program.  If not, see http://www.gnu.org/licenses.
#
# http://numenta.org/licenses/
# ----------------------------------------------------------------------

"""
## @file
Internal package.

Package containing modules that are used internally by Numenta Python
tools and plugins to extend standard library functionality.
These modules should NOT be used by client applications.

The following modules are included:

nupic.support.paths
Module containing filesystem path manipulation utilities.

nupic.support.serialization
Module containing Python object serialization (pickling and unpickling) and
versioning utilities.

nupic.support.compress
Module containing Python object encoding and compression utilities.

nupic.support.processes
Module containing operating system process management utilities and wrappers.

nupic.support.output
Module containing operating system interprocess communication utilities and
wrappers.

nupic.support.diff
Module containing file difference calculation wrappers.

nupic.support.vision
Temporary location for vision framework before the move to nupic.vision.

nupic.support.deprecate
Contains the deprecate decorator used for automatic handling of deprecated
methods.

nupic.support.memchecker
Contains the MemChecker class, for checking physical memory and monitoring
memory usage.

nupic.support.imagesearch
Contains functions for searching for images on the web and downloading them.
"""

from __future__ import with_statement

# Standard imports
import os
import sys
import inspect
import logging
import logging.config
import logging.handlers
from platform import python_version
import struct
from StringIO import StringIO
import time
import traceback

from pkg_resources import resource_string, resource_filename

from configuration import Configuration
from nupic.support.fshelpers import makeDirectoryFromAbsolutePath


# Local imports



def getCallerInfo(depth=2):
  """Utility function to get information about function callers

  The information is the tuple (function/method name, filename, class)
  The class will be None if the caller is just a function and not an object
  method.

  depth: how far back in the callstack to go to extract the caller info

  """
  f = sys._getframe(depth)
  method_name = f.f_code.co_name
  filename = f.f_code.co_filename

  arg_class = None
  args = inspect.getargvalues(f)
  if len(args[0]) > 0:
    arg_name = args[0][0] # potentially the 'self' arg if its a method
    arg_class = args[3][arg_name].__class__.__name__
  return (method_name, filename, arg_class)



def title(s=None, additional='', stream=sys.stdout, frame='-'):
  """Utility function to display nice titles

  It automatically extracts the name of the function/method it is called from
  and you can add additional text. title() will then print the name
  of the function/method and the additional text surrounded by tow lines
  of dashes. If you don't want the name of the function, you can provide
  alternative text (regardless of the additional text)

  @param s - text to display, uses the function name and arguments by default
  @param additional - extra text to display (not needed if s is not None)
  @param stream - the stream to print to. Ny default goes to standard output
  @param frame - the character used for the over and under line. Default is '-'

  Examples:

  def foo():
    title()

  will display:

  ---
  foo
  ---

  def foo():
    title(additional='(), this is cool!!!')

  will display:

  ----------------------
  foo(), this is cool!!!
  ----------------------


  def foo():
    title('No function name here!')

  will display:

  ----------------------
  No function name here!
  ----------------------
  """
  if s is None:
    callable_name, file_name, class_name = getCallerInfo(2)
    s = callable_name
    if class_name is not None:
      method_name = s
      s = class_name + '.' + callable_name
  lines = (s + additional).split('\n')
  length = max(len(line) for line in lines)
  print >> stream, '-' * length
  print >> stream, s + additional
  print >> stream, '-' * length



def bringToFront(title):
  """Bring a top-level window with a given title
     to the front on Windows"""
  if sys.platform != 'win32':
    return

  import ctypes
  find_window = ctypes.windll.user32.FindWindowA
  set_foreground_window = ctypes.windll.user32.SetForegroundWindow
  hwnd = find_window(None, title)
  if hwnd == 0:
    raise Exception('There is no window titled: "%s"' % title)
  set_foreground_window(hwnd)



def getUserDocumentsPath():
  """
  Find the user's "Documents" directory (OS X), "My Documents" directory
  (Windows), or home directory (Unix).
  """

  # OS X and Windows code from:
  # http://www.blueskyonmars.com/2005/08/05
  # /finding-a-users-my-documents-folder-on-windows/
  # Alternate Windows code from:
  # http://bugs.python.org/issue1763
  if sys.platform.startswith('win'):
    if sys.platform.startswith('win32'):
      # Try the primary method on 32-bit windows
      try:
        from win32com.shell import shell
        alt = False
      except ImportError:
        try:
          import ctypes
          dll = ctypes.windll.shell32
          alt = True
        except:
          raise Exception("Could not find 'My Documents'")
    else:
      # Use the alternate method on 64-bit Windows
      alt = True
    if not alt:
      # Primary method using win32com
      df = shell.SHGetDesktopFolder()
      pidl = df.ParseDisplayName(0, None,
               "::{450d8fba-ad25-11d0-98a8-0800361b1103}")[1]
      path = shell.SHGetPathFromIDList(pidl)
    else:
      # Alternate method using ctypes rather than win32com
      buf = ctypes.create_string_buffer(300)
      dll.SHGetSpecialFolderPathA(None, buf, 0x0005, False)
      path = buf.value
  elif sys.platform.startswith('darwin'):
    from Carbon import Folder, Folders
    folderref = Folder.FSFindFolder(Folders.kUserDomain,
                                    Folders.kDocumentsFolderType,
                                    False)
    path = folderref.as_pathname()
  else:
    path = os.getenv('HOME')
  return path



def getArgumentDescriptions(f):
  """
  Get the arguments, default values, and argument descriptions for a function.

  Returns a list of tuples: (argName, argDescription, defaultValue). If an
    argument has no default value, the tuple is only two elements long (as None
    cannot be used, since it could be a default value itself).

  Parses the argument descriptions out of the function docstring, using a
  format something lke this:

  [junk]
  argument_name:     description...
    description...
    description...
  [junk]
  [more arguments]

  It will find an argument as long as the exact argument name starts the line.
  It will then strip a trailing colon, if present, then strip the rest of the
  line and use it to start the description. It will then strip and append any
  subsequent lines with a greater indent level than the original argument name.
  """

  # Get the argument names and default values
  argspec = inspect.getargspec(f)

  # Scan through the docstring to extract documentation for each argument as
  # follows:
  #   Check the first word of the line, stripping a colon if one is present.
  #   If it matches an argument name:
  #    Take the rest of the line, stripping leading whitespeace
  #    Take each subsequent line if its indentation level is greater than the
  #      initial indentation level
  #    Once the indentation level is back to the original level, look for
  #      another argument
  docstring = f.__doc__
  descriptions = {}
  if docstring:
    lines = docstring.split('\n')
    i = 0
    while i < len(lines):
      stripped = lines[i].lstrip()
      if not stripped:
        i += 1
        continue
      # Indentation level is index of the first character
      indentLevel = lines[i].index(stripped[0])
      # Get the first word and remove the colon, if present
      firstWord = stripped.split()[0]
      if firstWord.endswith(':'):
        firstWord = firstWord[:-1]
      if firstWord in argspec.args:
        # Found an argument
        argName = firstWord
        restOfLine = stripped[len(firstWord)+1:].strip()
        argLines = [restOfLine]
        # Take the next lines as long as they are indented more
        i += 1
        while i < len(lines):
          stripped = lines[i].lstrip()
          if not stripped:
            # Empty line - stop
            break
          if lines[i].index(stripped[0]) <= indentLevel:
            # No longer indented far enough - stop
            break
          # This line counts too
          argLines.append(lines[i].strip())
          i += 1
        # Store this description
        descriptions[argName] = ' '.join(argLines)
      else:
        # Not an argument
        i += 1

  # Build the list of (argName, description, defaultValue)
  args = []
  if argspec.defaults:
    defaultCount = len(argspec.defaults)
  else:
    defaultCount = 0
  nonDefaultArgCount = len(argspec.args) - defaultCount
  for i, argName in enumerate(argspec.args):
    if i >= nonDefaultArgCount:
      defaultValue = argspec.defaults[i - nonDefaultArgCount]
      args.append((argName, descriptions.get(argName, ""), defaultValue))
    else:
      args.append((argName, descriptions.get(argName, "")))

  return args



# TODO queryNumInwardIters appears to be unused and should probably be deleted
#  from here altogether; it's likely an artifact of the legacy vision support.
#def queryNumInwardIters(configPath, radialLength, numRepetitions=1):
#  """
#  Public utility API that accepts a config path and
#  radial length, and determines the proper number of
#  training iterations with which to invoke net.run()
#  when running a PictureSensor in 'inward' mode.
#  """
#  numCats = queryNumCategories(configPath)
#  sequenceLen = radialLength + 1
#  numItersPerCat = (8 * radialLength) * sequenceLen
#  numTrainingItersTP = numItersPerCat * numCats
#  return numTrainingItersTP * numRepetitions



gLoggingInitialized = False
def initLogging(verbose=False, console='stdout', consoleLevel='DEBUG'):
  """
  Initilize NuPic logging by reading in from the logging configuration file. The
  logging configuration file is named 'nupic-logging.conf' and is expected to be
  in the format defined by the python logging module.

  If the environment variable 'NTA_CONF_PATH' is defined, then the logging
  configuration file is expected to be in the NTA_CONF_PATH directory. If
  NTA_CONF_PATH is not defined, then it is found in the 'conf/default'
  subdirectory of the NuPic installation directory (typically
  ~/nupic/current/conf/default)

  The logging configuration file can use the environment variable 'NTA_LOG_DIR'
  to set the locations of log files. If this variable is not defined, logging to
  files will be disabled.
  
  console:    Defines console output for the default "root" logging
              configuration; this may be one of 'stdout', 'stderr', or None;
              Use None to suppress console logging output
  consoleLevel:
              Logging-level filter string for console output corresponding to
              logging levels in the logging module; may be one of:
              'DEBUG', 'INFO', 'WARNING', 'ERROR', or 'CRITICAL'.
              E.g.,  a value of'WARNING' suppresses DEBUG and INFO level output
              to console, but allows WARNING, ERROR, and CRITICAL 
  """

  # NOTE: If you call this twice from the same process there seems to be a
  # bug - logged messages don't show up for loggers that you do another
  # logging.getLogger() on.
  global gLoggingInitialized
  if gLoggingInitialized:
    if verbose:
      print >> sys.stderr, "Logging already initialized, doing nothing."
    return
  
  consoleStreamMappings = {
    'stdout'  : 'stdoutConsoleHandler',
    'stderr'  : 'stderrConsoleHandler',
  }
  
  consoleLogLevels = ['DEBUG', 'INFO', 'WARNING', 'WARN', 'ERROR', 'CRITICAL',
                      'FATAL']
  
  assert console is None or console in consoleStreamMappings.keys(), (
    'Unexpected console arg value: %r') % (console,)
  
  assert consoleLevel in consoleLogLevels, (
    'Unexpected consoleLevel arg value: %r') % (consoleLevel)

  # -----------------------------------------------------------------------
  # Setup logging. Look for the nupic-logging.conf file, first in the
  #   NTA_CONFIG_DIR path (if defined), then in a subdirectory of the nupic
  #   module
  configFilename = 'nupic-logging.conf'
  configFilePath = resource_filename("nupic.support", configFilename)
  
  configLogDir = os.environ.get('NTA_LOG_DIR', None)

  # Load in the logging configuration file
  if verbose:
    print >> sys.stderr, (
      "Using logging configuration file: %s") % (configFilePath)

  # This dict will hold our replacement strings for logging configuration
  replacements = dict()

  def makeKey(name):
    """ Makes replacement key """
    return "$$%s$$" % (name)

  platform = sys.platform.lower()
  if platform.startswith('java'):
    # Jython
    import java.lang
    platform = java.lang.System.getProperty("os.name").lower()
    if platform.startswith('mac os x'):
      platform = 'darwin'

  if platform.startswith('darwin'):
    replacements[makeKey('SYSLOG_HANDLER_ADDRESS')] = '"/var/run/syslog"'
  elif platform.startswith('linux'):
    replacements[makeKey('SYSLOG_HANDLER_ADDRESS')] = '"/dev/log"'
  elif platform.startswith('win'):
    replacements[makeKey('SYSLOG_HANDLER_ADDRESS')] = '"log"'
  else:
    raise RuntimeError("This platform is neither darwin, win32, nor linux: %s" % (
      sys.platform,))

  # Nupic logs go to file
  replacements[makeKey('PERSISTENT_LOG_HANDLER')] = 'fileHandler'
  if platform.startswith('win'):
    replacements[makeKey('FILE_HANDLER_LOG_FILENAME')] = '"NUL"'
  else:
    replacements[makeKey('FILE_HANDLER_LOG_FILENAME')] = '"/dev/null"'

  # Set up log file path for the default file handler and configure handlers
  handlers = list()
  
  if configLogDir is not None:
    logFilePath = _genLoggingFilePath()
    makeDirectoryFromAbsolutePath(os.path.dirname(logFilePath))
    replacements[makeKey('FILE_HANDLER_LOG_FILENAME')] = repr(logFilePath)

    handlers.append(replacements[makeKey('PERSISTENT_LOG_HANDLER')])

  if console is not None:
    handlers.append(consoleStreamMappings[console])

  replacements[makeKey('ROOT_LOGGER_HANDLERS')] = ", ".join(handlers)

  # Set up log level for console handlers
  replacements[makeKey('CONSOLE_LOG_LEVEL')] = consoleLevel

  customConfig = StringIO()

  # Using pkg_resources to get the logging file, which should be packaged and
  # associated with this source file name.
  loggingFileContents = resource_string(__name__, configFilename)

  for lineNum, line in enumerate(loggingFileContents.splitlines()):
    if "$$" in line:
      for (key, value) in replacements.items():
        line = line.replace(key, value)

    # If there is still a replacement string in the line, we're missing it
    #  from our replacements dict
    if "$$" in line and "$$<key>$$" not in line:
      raise RuntimeError(("The text %r, found at line #%d of file %r, "
                          "contains a string not found in our replacement "
                          "dict.") % (line, lineNum, configFilePath))

    customConfig.write("%s\n" % line)

  customConfig.seek(0)
  if python_version()[:3] >= '2.6':
    logging.config.fileConfig(customConfig, disable_existing_loggers=False)
  else:
    logging.config.fileConfig(customConfig)

  gLoggingInitialized = True



def reinitLoggingDir():
  """ (Re-)Initialize the loging directory for the calling application that
  uses initLogging() for logging configuration
  
  NOTE: It's typially unnecessary to call this function directly since
   initLogging takes care of it for you. This function is exposed primarily for
   the benefit of nupic-services.py to allow it to restore its logging directory
   after the hard-reset operation.
  """
  if gLoggingInitialized and 'NTA_LOG_DIR' in os.environ:
    makeDirectoryFromAbsolutePath(os.path.dirname(_genLoggingFilePath()))



def _genLoggingFilePath():
  """ Generate a filepath for the calling app """
  appName = os.path.splitext(os.path.basename(sys.argv[0]))[0] or 'UnknownApp'
  appLogDir = os.path.abspath(os.path.join(
    os.environ['NTA_LOG_DIR'],
    'numenta-logs-%s' % (os.environ['USER'],),
    appName))
  appLogFileName = '%s-%s-%s.log' % (
    appName, long(time.mktime(time.gmtime())), os.getpid())
  return os.path.join(appLogDir, appLogFileName)
  
  

def enableLoggingErrorDebugging():
  """ Overrides the python logging facility's Handler.handleError function to
  raise an exception instead of print and suppressing it.  This allows a deeper
  stacktrace to be emitted that is very helpful for quickly finding the
  file/line that initiated the invalidly-formatted logging operation.
  
  NOTE: This is for debugging only - be sure to remove the call to this function
   *before* checking in your changes to the source code repository, as it will
   cause the application to fail if some invalidly-formatted logging statement
   still exists in your code.
  
  Example usage: enableLoggingErrorDebugging must be called *after*
   initLogging()
   
    import nupic.support
    nupic.support.initLogging()
    nupic.support.enableLoggingErrorDebugging()
  
  "TypeError: not all arguments converted during string formatting" is an
  example exception that might be output by the built-in handlers with the
  following very shallow traceback that doesn't go deep enough to show the
  source of the problem:
  
  File ".../python2.6/logging/__init__.py", line 776, in emit
    msg = self.format(record)
  File ".../python2.6/logging/__init__.py", line 654, in format
    return fmt.format(record)
  File ".../python2.6/logging/__init__.py", line 436, in format
    record.message = record.getMessage()
  File ".../python2.6/logging/__init__.py", line 306, in getMessage
    msg = msg % self.args
  TypeError: not all arguments converted during string formatting
  """
  
  print >> sys.stderr, ("WARNING")
  print >> sys.stderr, ("WARNING: "
    "nupic.support.enableLoggingErrorDebugging() was "
    "called to install a debugging patch into all logging handlers that "
    "will cause the program to fail if a logging exception occurrs; this "
    "call is for debugging only and MUST be removed before checking in code "
    "into production system. Caller: %s") % (
    traceback.format_stack(),)
  print >> sys.stderr, ("WARNING")
  
  def handleErrorPatch(*args, **kwargs):
    if logging.raiseExceptions:
      raise
  
  for handler in logging._handlerList:
    handler.handleError = handleErrorPatch
  
  return



def intTo8ByteArray(inValue):
  """
  Converts an int to a packed byte array, with left most significant byte
  """

  values = (
    (inValue >> 56 ) & 0xff,
    (inValue >> 48 ) & 0xff,
    (inValue >> 40 ) & 0xff,
    (inValue >> 32 ) & 0xff,
    (inValue >> 24 ) & 0xff,
    (inValue >> 16 ) & 0xff,
    (inValue >> 8 ) & 0xff,
    inValue & 0xff
  )

  s = struct.Struct('B B B B B B B B')
  packed_data = s.pack(*values)

  return packed_data



def byteArrayToInt(packed_data):
  """
  Converts a byte array into an integer
  """
  value = struct.unpack('B B B B B B B B', packed_data)
  return value[0] << 56 | \
         value[1] << 48 | \
         value[2] << 40 | \
         value[3] << 32 | \
         value[4] << 24 | \
         value[5] << 16 | \
         value[6] << 8 | \
         value[7]



def getSpecialRowID():
  """
  Special row id is 0xFF FFFF FFFF FFFF FFFF (9 bytes of 0xFF)
  """
  values = (0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF)
  s = struct.Struct('B B B B B B B B B')
  packed_data = s.pack(*values)

  return packed_data



_FLOAT_SECONDS_IN_A_DAY = 24.0 * 60.0 * 60.0
def floatSecondsFromTimedelta(td):
  """ Convert datetime.timedelta to seconds in floating point """
  sec = (td.days * _FLOAT_SECONDS_IN_A_DAY + td.seconds * 1.0 +
         td.microseconds / 1E6)

  return sec



def aggregationToMonthsSeconds(interval):
  """
  Return the number of months and seconds from an aggregation dict that 
  represents a date and time. 
  
  Interval is a dict that contain one or more of the following keys: 'years',
  'months', 'weeks', 'days', 'hours', 'minutes', seconds', 'milliseconds', 
  'microseconds'.

  Parameters:
  ---------------------------------------------------------------------
  interval:  The aggregation interval, as a dict representing a date and time
  retval:    number of months and seconds in the interval, as a dict:
                {months': XX, 'seconds': XX}. The seconds is
                a floating point that can represent resolutions down to a
                microsecond. 
  
  For example:
  aggregationMicroseconds({'years': 1, 'hours': 4, 'microseconds':42}) == 
      {'months':12, 'seconds':14400.000042}
  
  """
  
  seconds = interval.get('microseconds', 0) * 0.000001
  seconds += interval.get('milliseconds', 0) * 0.001
  seconds += interval.get('seconds', 0) 
  seconds += interval.get('minutes', 0) * 60
  seconds += interval.get('hours', 0) * 60 * 60
  seconds += interval.get('days', 0) * 24 * 60 * 60
  seconds += interval.get('weeks', 0) * 7 * 24 * 60 * 60
  
  months = interval.get('months', 0)
  months += 12 * interval.get('years', 0)
  
  return {'months': months, 'seconds': seconds}



def aggregationDivide(dividend, divisor):
  """
  Return the result from dividing two dicts that represent date and time.
  
  Both dividend and divisor are dicts that contain one or more of the following
  keys: 'years', 'months', 'weeks', 'days', 'hours', 'minutes', seconds',
  'milliseconds', 'microseconds'.
  
  Parameters:
  ---------------------------------------------------------------------
  dividend:  The numerator, as a dict representing a date and time
  divisor:   the denominator, as a dict representing a date and time
  retval:    number of times divisor goes into dividend, as a floating point
                number. 
  
  For example:
  aggregationDivide({'hours': 4}, {'minutes': 15}) == 16
  
  """
  
  # Convert each into microseconds
  dividendMonthSec = aggregationToMonthsSeconds(dividend)
  divisorMonthSec = aggregationToMonthsSeconds(divisor)
  
  # It is a usage error to mix both months and seconds in the same operation
  if (dividendMonthSec['months'] != 0 and divisorMonthSec['seconds'] != 0) \
    or (dividendMonthSec['seconds'] != 0 and divisorMonthSec['months'] != 0):
    raise RuntimeError("Aggregation dicts with months/years can only be "
      "inter-operated with other aggregation dicts that contain "
      "months/years")
    
  
  if dividendMonthSec['months'] > 0:
    return float(dividendMonthSec['months']) / divisor['months']
  
  else:
    return float(dividendMonthSec['seconds']) / divisorMonthSec['seconds']



#! /usr/bin/env python
# ----------------------------------------------------------------------
# Numenta Platform for Intelligent Computing (NuPIC)
# Copyright (C) 2013, Numenta, Inc.  Unless you have an agreement
# with Numenta, Inc., for a separate license for this software code, the
# following terms and conditions apply:
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero Public License for more details.
#
# You should have received a copy of the GNU Affero Public License
# along with this program.  If not, see http://www.gnu.org/licenses.
#
# http://numenta.org/licenses/
# ----------------------------------------------------------------------


'''
This module implements the Feature Flags system.

Summary:

  Based on the environment your program is run in you may want various features
  to be displayed and others not.

  Based on the type of user who is using Nupic you may want them to see various
  features and not others.

  These two things can be accomplished using Feature Flags.

  Feature Flags wrap sections of code in IF/THEN blocks to isolate functionality
  from running code.

  This allows you to inline new features without branching files into file.py and
  file2.py.

  This also allows you to release un-tested, underdevelopment, or experimental
  features directly to production with confidence that those code paths won't
  get run until such time as they are 'released' to a group of users for testing
  or to all users for a launch.

  DEV NOTE: To set personal Feature Flag over-rides copy conf/devconf-example.py
            to devconf.py and edit as desired.

'''
import os
import sys

from nupic.support.features_list import FEATURES_LIST
from nupic.support.feature_groups import GROUPS


class Features(object):
  '''
  This class can be used to retrieve available and calculated features for
  a given combination of running environment and user

  The class is for documentation purposes as all the methods are statically
  accessible.

  Modeled after Ron's excellent Configuration class
  '''


  @staticmethod
  def hasFeature(feature, group = False, addList = False, removeList = False):
    '''
    This is the primary method of the class that will return True or False
    based on the current environment and user.

    This is the 'Flag' part of Feature Flags.
    '''

    return (feature in Features.getFeatures(group, addList, removeList))

  @staticmethod
  def getFeatures(group = False, addList = False, removeList = False):
    '''
    Returns a list of all the active features for the current env/user

    groups:         This is a list of groups the current user belongs to
    addList:        A list of one-off features to add
    removeList:     A list of one-off features to remove
    '''

    # Calculate groups
    if group:
      featureList = Features._getFeaturesForGroup(group)
    else:
      featureList = Features._getFeaturesForGroup('BASE')

    # Calculate user one-offs
    if addList:
      featureList.extend(addList)
    if removeList:
      for item in removeList:
        if item in featureList:
          featureList.remove(item)

    # Caclulate developer configuration
    if 'TRUNK' in os.environ:
      confDir = os.path.join(os.environ['TRUNK'], 'config')
      confFilepath = os.path.join(confDir, 'developer.py')
      if os.path.exists(confFilepath):
        sys.path.append(confDir)
        import developer
        featureList.extend(developer.FEATURES['ADD'])
        for item in developer.FEATURES['REMOVE']:
          if item in featureList:
            featureList.remove(item)

    return featureList

  @staticmethod
  def _getFeaturesForGroup(group):
    '''
    Using feature_groups.py calculate the set of features available to the given
    group
    '''
    return GROUPS[group]

  @staticmethod
  def getAllFeatures():
    '''
    Returns a list of all known features (essentially the contents of feature_list.py)
    '''
    for feature in FEATURES_LIST:
      print feature['name'] + '\t\t' + feature['description']

  @staticmethod
  def getAllGroups():
    '''
    Returns a list of all known feature groups
    '''
    return [group for group, features in GROUPS.iteritems()]



#!/usr/bin/env python
# ----------------------------------------------------------------------
# Numenta Platform for Intelligent Computing (NuPIC)
# Copyright (C) 2013, Numenta, Inc.  Unless you have an agreement
# with Numenta, Inc., for a separate license for this software code, the
# following terms and conditions apply:
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero Public License for more details.
#
# You should have received a copy of the GNU Affero Public License
# along with this program.  If not, see http://www.gnu.org/licenses.
#
# http://numenta.org/licenses/
# ----------------------------------------------------------------------

"""Tool for adding anomalies to data sets."""

import random
import sys

from nupic.data.file_record_stream import FileRecordStream

USAGE = """
Usage:
  python anomalyzer.py input output action extraArgs
Actions:
  add
    extraArgs: column start stop value
    Adds value to each element in range [start, stop].
  scale
    extraArgs: column start stop multiple
    Multiplies each element in range [start, stop] by multiple.
  copy
    extraArgs: start stop [insertLocation] [tsCol]
    Copies the values in range [start, stop] and inserts them at
    insertLocation, or following the copied section if no insertLocation
    is specified. Updates timestamps if tsCol is given.
  sample
    extraArgs: n [start] [stop] [tsCol]
    Samples rows from the specified range and writes them to the new file.
    If tsCol is specified, the timestamps of this column are updated.
  sample2
    Same as sample except the rows before and after the specified range are
    preserved.
"""



class Actions(object):
  """Enum class for actions that can be performed."""
  ADD = 'add'
  SCALE = 'scale'
  COPY = 'copy'
  SAMPLE = 'sample'
  SAMPLE2 = 'sample2'
  ACTIONS = (ADD, SCALE, COPY, SAMPLE, SAMPLE2)



def add(reader, writer, column, start, stop, value):
  """Adds a value over a range of rows.

  Args:
    reader: A FileRecordStream object with input data.
    writer: A FileRecordStream object to write output data to.
    column: The column of data to modify.
    start: The first row in the range to modify.
    end: The last row in the range to modify.
    value: The value to add.
  """
  for i, row in enumerate(reader):
    if i >= start and i <= stop:
      row[column] = type(value)(row[column]) + value
    writer.appendRecord(row)



def scale(reader, writer, column, start, stop, multiple):
  """Multiplies a value over a range of rows.

  Args:
    reader: A FileRecordStream object with input data.
    writer: A FileRecordStream  object to write output data to.
    column: The column of data to modify.
    start: The first row in the range to modify.
    end: The last row in the range to modify.
    multiple: The value to scale/multiply by.
  """
  for i, row in enumerate(reader):
    if i >= start and i <= stop:
      row[column] = type(multiple)(row[column]) * multiple
    writer.appendRecord(row)



def copy(reader, writer, start, stop, insertLocation=None, tsCol=None):
  """Copies a range of values to a new location in the data set.

  Args:
    reader: A FileRecordStream object with input data.
    writer: A FileRecordStream object to write output data to.
    start: The first row in the range to copy.
    stop: The last row in the range to copy.
    insertLocation: The location to insert the copied range. If not specified,
        the range is inserted immediately following itself.
  """
  assert stop >= start
  startRows = []
  copyRows = []
  ts = None
  inc = None
  if tsCol is None:
    tsCol = reader.getTimestampFieldIdx()
  for i, row in enumerate(reader):
    # Get the first timestamp and the increment.
    if ts is None:
      ts = row[tsCol]
    elif inc is None:
      inc = row[tsCol] - ts
    # Keep a list of all rows and a list of rows to copy.
    if i >= start and i <= stop:
      copyRows.append(row)
    startRows.append(row)
  # Insert the copied rows.
  if insertLocation is None:
    insertLocation = stop + 1
  startRows[insertLocation:insertLocation] = copyRows
  # Update the timestamps.
  for row in startRows:
    row[tsCol] = ts
    writer.appendRecord(row)
    ts += inc



def sample(reader, writer, n, start=None, stop=None, tsCol=None,
           writeSampleOnly=True):
  """Samples n rows.

  Args:
    reader: A FileRecordStream object with input data.
    writer: A FileRecordStream object to write output data to.
    n: The number of elements to sample.
    start: The first row in the range to sample from.
    stop: The last row in the range to sample from.
    tsCol: If specified, the timestamp column to update.
    writeSampleOnly: If False, the rows before start are written before the
        sample and the rows after stop are written after the sample.
  """
  rows = list(reader)
  if tsCol is not None:
    ts = rows[0][tsCol]
    inc = rows[1][tsCol] - ts
  if start is None:
    start = 0
  if stop is None:
    stop = len(rows) - 1
  initialN = stop - start + 1
  # Select random rows in the sample range to delete until the desired number
  # of rows are left.
  numDeletes =  initialN - n
  for i in xrange(numDeletes):
    delIndex = random.randint(start, stop - i)
    del rows[delIndex]
  # Remove outside rows if specified.
  if writeSampleOnly:
    rows = rows[start:start + n]
  # Rewrite columns if tsCol is given.
  if tsCol is not None:
    ts = rows[0][tsCol]
  # Write resulting rows.
  for row in rows:
    if tsCol is not None:
      row[tsCol] = ts
      ts += inc
    writer.appendRecord(row)



def main(args):
  inputPath, outputPath, action = args[:3]
  with FileRecordStream(inputPath) as reader:
    with FileRecordStream(outputPath, write=True,
                          fields=reader.fields) as writer:
      assert action in Actions.ACTIONS, USAGE
      if action == Actions.ADD:
        assert len(args) == 7, USAGE
        start = int(args[4])
        stop = int(args[5])
        column = int(args[3])
        valueType = eval(reader.fields[column][1])
        value = valueType(args[6])
        add(reader, writer, column, start, stop, value)
      elif action == Actions.SCALE:
        assert len(args) == 7, USAGE
        start = int(args[4])
        stop = int(args[5])
        column = int(args[3])
        valueType = eval(reader.fields[column][1])
        multiple = valueType(args[6])
        scale(reader, writer, column, start, stop, multiple)
      elif action == Actions.COPY:
        assert 5 <= len(args) <= 8, USAGE
        start = int(args[3])
        stop = int(args[4])
        if len(args) > 5:
          insertLocation = int(args[5])
        else:
          insertLocation = None
        if len(args) == 7:
          tsCol = int(args[6])
        else:
          tsCol = None
        copy(reader, writer, start, stop, insertLocation, tsCol)
      elif action == Actions.SAMPLE or action == Actions.SAMPLE2:
        assert 4 <= len(args) <= 7, USAGE
        n = int(args[3])
        start = None
        if len(args) > 4:
          start = int(args[4])
        stop = None
        if len(args) > 5:
          stop = int(args[5])
        tsCol = None
        if len(args) > 6:
          tsCol = int(args[6])
        writeSampleOnly = action == Actions.SAMPLE
        sample(reader, writer, n, start, stop, tsCol, writeSampleOnly)



if __name__ == "__main__":
  if len(sys.argv) <= 1:
    print USAGE
    sys.exit(1)
  main(sys.argv[1:])



# ----------------------------------------------------------------------
# Numenta Platform for Intelligent Computing (NuPIC)
# Copyright (C) 2013, Numenta, Inc.  Unless you have an agreement
# with Numenta, Inc., for a separate license for this software code, the
# following terms and conditions apply:
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero Public License for more details.
#
# You should have received a copy of the GNU Affero Public License
# along with this program.  If not, see http://www.gnu.org/licenses.
#
# http://numenta.org/licenses/
# ----------------------------------------------------------------------
import os

# This is the environment variable that controls the lock attributes
# enforcement.
#
# The lock attributes machinery is engaged by default. To deactivate it
# define this environment variabe. The benefit is that there will be no runtime
# overhead (Except for a one-time check when classes that derive from
# LockAttributesMixin are defined or methods decorated with
# _canAddAttributes are defined).
deactivation_key = 'NTA_DONT_USE_LOCK_ATTRIBUTES'

def _allow_new_attributes(f):
  """A decorator that maintains the attribute lock state of an object

  It coperates with the LockAttributesMetaclass (see bellow) that replaces
  the __setattr__ method with a custom one that checks the _canAddAttributes
  counter and allows setting new attributes only if _canAddAttributes > 0.

  New attributes can be set only from methods decorated
  with this decorator (should be only __init__ and __setstate__ normally)

  The decorator is reentrant (e.g. if from inside a decorated function another
  decorated function is invoked). Before invoking the target function it
  increments the counter (or sets it to 1). After invoking the target function
  it decrements the counter and if it's 0 it removed the counter.
  """
  def decorated(self, *args, **kw):
    """The decorated function that replaces __init__() or __setstate__()

    """
    # Run the original function
    if not hasattr(self, '_canAddAttributes'):
      self.__dict__['_canAddAttributes'] = 1
    else:
      self._canAddAttributes += 1
    assert self._canAddAttributes >= 1

    # Save add attribute counter
    count = self._canAddAttributes
    f(self, *args, **kw)

    # Restore _CanAddAttributes if deleted from dict (can happen in __setstte__)
    if hasattr(self, '_canAddAttributes'):
      self._canAddAttributes -= 1
    else:
      self._canAddAttributes = count - 1

    assert self._canAddAttributes >= 0
    if self._canAddAttributes == 0:
      del self._canAddAttributes

  decorated.__doc__ = f.__doc__
  decorated.__name__ = f.__name__
  return decorated

def _simple_init(self, *args, **kw):
  """trivial init method that just calls base class's __init__()

  This method is attached to classes that don't define __init__(). It is needed
  because LockAttributesMetaclass must decorate the __init__() method of
  its target class.
  """
  type(self).__base__.__init__(self, *args, **kw)

class LockAttributesMetaclass(type):
  """This metaclass makes objects attribute-locked by decorating their
  __init__() and __setstate__() methods with the _allow_new_attributes
  decorator.

  It doesn't do anything unless the environment variable
  'NTA_USE_LOCK_ATTRIBUTES' is defined.
  That allows for verifying proper usage during testing and skipping
  it in production code (that was verified during testing) to avoid the cost
  of verifying every attribute setting.

  It also replaces the __setattr__ magic method with a custom one that verifies
  new attributes are set only in code that originates from a decorated method
  (normally __init__() or __setstate__()).

  If the target class has no __init__() method it adds a trivial __init__()
  method to provide a hook for the decorator (the _simple_init()
  function defined above)
  """
  def __init__(cls, name, bases, dict):
    """
    """
    def custom_setattr(self, name, value):
      """A custom replacement for __setattr__

      Allows setting only exisitng attributes. It is designed to work
      with the _allow_new_attributes decorator.

      It works is by checking if the requested attribute is already in the
      __dict__ or if the _canAddAttributes counter > 0. Otherwise it raises an
      exception.
      If all is well it calls the original __setattr__. This means it can work
      also with classes that already have custom __setattr__
      """
      if (name == '_canAddAttributes' or
         (hasattr(self, '_canAddAttributes') and self._canAddAttributes > 0) or
         hasattr(self, name)):
        return self._original_setattr(name, value)
      else:
        #from dbgp.client import brk; brk(port=9029)
        raise Exception('Attempting to set a new attribute: ' + name)


    # Bail out if not active. Zero overhead other than this one-time check
    # at class definition time
    if deactivation_key in os.environ:
      return

    # Initialize the super-class
    super(LockAttributesMetaclass, cls).__init__(name, bases, dict)

    # Store and replace the __setattr__ with the custom one (if needed)
    if not hasattr(cls, '_original_setattr'):
      cls._original_setattr = cls.__setattr__
      cls.__setattr__ = custom_setattr


    # Keep the original __init__ if exists. This was needed for NuPIC 1. Remove?
    if '__init__' in dict:
      setattr(cls, '_original_init', dict['__init__'])


    # Get the __init__ and __setstate__ form the target class's dict
    # If there is no __init__ use _simple_init (it's Ok if there is no
    #__setstate__)
    methods = [('__init__', dict.get('__init__', _simple_init)),
               ('__setstate__', dict.get('__setstate__', None))]

    # Wrap the methods with _allow_new_attributes decorator
    for name, method in methods:
      if method is not None:
        setattr(cls, name, _allow_new_attributes(method))

class LockAttributesMixin(object):
  """This class serves as a base (or mixin) for classes that want to enforce
  the locked attributes pattern (all attributes should be defined in __init__()
  or __setstate__().

  All the target class has to do add LockAttributesMixin as one of its bases
  (inherit from it).

  The metaclass will be activated when the application class is created
  and the lock attributes machinery will be injected (unless the
  deactivation_key is defined in the environment)
  """
  __metaclass__ = LockAttributesMetaclass



# ----------------------------------------------------------------------
# Numenta Platform for Intelligent Computing (NuPIC)
# Copyright (C) 2013, Numenta, Inc.  Unless you have an agreement
# with Numenta, Inc., for a separate license for this software code, the
# following terms and conditions apply:
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero Public License for more details.
#
# You should have received a copy of the GNU Affero Public License
# along with this program.  If not, see http://www.gnu.org/licenses.
#
# http://numenta.org/licenses/
# ----------------------------------------------------------------------


# This script contains file-system helper functions


import os



def makeDirectoryFromAbsolutePath(absDirPath):
  """ Makes directory for the given directory path with default permissions.
  If the directory already exists, it is treated as success.

  absDirPath:   absolute path of the directory to create.

  Returns:      absDirPath arg

  Exceptions:         OSError if directory creation fails
  """

  assert os.path.isabs(absDirPath)

  try:
    os.makedirs(absDirPath)
  except OSError, e:
    if e.errno != os.errno.EEXIST:
      raise

  return absDirPath



# ----------------------------------------------------------------------
# Numenta Platform for Intelligent Computing (NuPIC)
# Copyright (C) 2014, Numenta, Inc.  Unless you have an agreement
# with Numenta, Inc., for a separate license for this software code, the
# following terms and conditions apply:
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero Public License for more details.
#
# You should have received a copy of the GNU Affero Public License
# along with this program.  If not, see http://www.gnu.org/licenses.
#
# http://numenta.org/licenses/
# ----------------------------------------------------------------------

"""
A shim for the TemporalMemory class that transparently implements TP,
for use with tests.
"""

import numpy

from nupic.math import GetNTAReal
from nupic.research.TP import TP
from nupic.research.TP10X2 import TP10X2
from nupic.research.connections import Connections



TPClass = TP10X2
dtype = GetNTAReal()



class TemporalMemoryShim(TPClass):
  """
  Temporal Memory => TP shim class.
  """
  def __init__(self,
               columnDimensions=(2048,),
               cellsPerColumn=32,
               activationThreshold=13,
               initialPermanence=0.21,
               connectedPermanence=0.50,
               minThreshold=10,
               maxNewSynapseCount=20,
               permanenceIncrement=0.10,
               permanenceDecrement=0.10,
               seed=42):
    """
    Translate parameters and initialize member variables
    specific to TemporalMemory
    """
    numberOfCols = 1
    for n in columnDimensions:
      numberOfCols *= n

    super(TemporalMemoryShim, self).__init__(
      numberOfCols=numberOfCols,
      cellsPerColumn=cellsPerColumn,
      initialPerm=initialPermanence,
      connectedPerm=connectedPermanence,
      minThreshold=minThreshold,
      newSynapseCount=maxNewSynapseCount,
      permanenceInc=permanenceIncrement,
      permanenceDec=permanenceDecrement,
      permanenceMax=1.0,
      globalDecay=0,
      activationThreshold=activationThreshold,
      seed=seed)

    self.connections = Connections(numberOfCols * cellsPerColumn)
    self.predictiveCells = set()


  def compute(self, activeColumns, learn=True):
    """
    Feeds input record through TM, performing inference and learning.
    Updates member variables with new state.

    @param activeColumns (set) Indices of active columns in `t`
    """
    bottomUpInput = numpy.zeros(self.numberOfCols, dtype=dtype)
    bottomUpInput[list(activeColumns)] = 1
    super(TemporalMemoryShim, self).compute(bottomUpInput,
                                            enableLearn=learn,
                                            computeInfOutput=True)

    predictedState = self.getPredictedState()
    self.predictiveCells = set(numpy.flatnonzero(predictedState))
